import logging
import typing

import dateutil.parser as dp

import ccc.elasticsearch
import cfg_mgmt.metrics
import cfg_mgmt.model as cmm
import cfg_mgmt.util as cmu
import ci.log
import ci.util
import model


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


def evaluate_cfg_element_status(
    cfg_element_status: cmm.CfgElementStatusReport,
) -> cmm.CfgStatusEvaluationResult:

    fully_compliant = True
    has_responsible = True
    has_rule = True
    assigned_rule_refers_to_undefined_policy = False
    has_status = True
    requires_status = None
    credentials_outdated = None
    non_compliant_reasons = []

    if not cfg_element_status.responsible:
        fully_compliant = False
        has_responsible = False
        non_compliant_reasons.append(cmm.CfgElementPolicyViolation.NO_RESPONSIBLE)

    if not cfg_element_status.rule:
        fully_compliant = False
        has_rule = False
        non_compliant_reasons.append(cmm.CfgElementPolicyViolation.NO_RULE)

    elif not cfg_element_status.policy:
        fully_compliant = False
        assigned_rule_refers_to_undefined_policy = True
        non_compliant_reasons.append(
            cmm.CfgElementPolicyViolation.ASSIGNED_RULE_REFERS_TO_UNDEFINED_POLICY
        )

    elif cfg_element_status.policy.type is cmm.PolicyType.MAX_AGE:
        policy = cfg_element_status.policy

        # status is only required if policy requires rotation
        if policy.max_age is None:
            requires_status = False
        else:
            requires_status = True

        if requires_status:
            if not (status := cfg_element_status.status):
                fully_compliant = False
                has_status = False
                non_compliant_reasons.append(cmm.CfgElementPolicyViolation.NO_STATUS)

            else:
                last_update = dp.isoparse(status.credential_update_timestamp)

                if policy.check(last_update=last_update, honour_grace_period=True):
                    credentials_outdated = False
                else:
                    fully_compliant = False
                    credentials_outdated = True
                    non_compliant_reasons.append(cmm.CfgElementPolicyViolation.CREDENTIALS_OUTDATED)

    else:
        raise NotImplementedError(cfg_element_status.policy.type)

    return cmm.CfgStatusEvaluationResult(
        fullyCompliant=fully_compliant,
        hasResponsible=has_responsible,
        hasRule=has_rule,
        assignedRuleRefersToUndefinedPolicy=assigned_rule_refers_to_undefined_policy,
        hasStatus=has_status,
        requiresStatus=requires_status,
        credentialsOutdated=credentials_outdated,
        nonCompliantReasons=non_compliant_reasons,
    )


def cfg_element_statuses_responsible_summaries(
    cfg_element_statuses: typing.Iterable[cmm.CfgElementStatusReport],
) -> typing.Generator[cmm.CfgResponsibleSummary, None, None]:

    responsible_summaries = dict()

    def responsible_summary(responsible: cmm.CfgResponsible, url: str) -> cmm.CfgResponsibleSummary:
        if (summary := responsible_summaries.get(responsible)):
            return summary

        cfg_responsible_summary = cmm.CfgResponsibleSummary(
            url=url,
            responsible=responsible,
            compliantElementsCount=0,
            noncompliantElementsCount=0,
        )
        responsible_summaries[responsible] = cfg_responsible_summary
        return cfg_responsible_summary

    for cfg_element_status in cfg_element_statuses:
        local_responsible_summaries: typing.List[cmm.CfgResponsibleSummary] = []

        if not cfg_element_status.responsible:
            continue

        for responsible in cfg_element_status.responsible.responsibles:
            cfg_responsible_summary = responsible_summary(
                responsible=responsible,
                url=cfg_element_status.element_storage,
            )
            local_responsible_summaries.append(cfg_responsible_summary)

        evaluation_result = evaluate_cfg_element_status(cfg_element_status)

        for summary in local_responsible_summaries:
            if evaluation_result.fullyCompliant:
                summary.compliantElementsCount += 1
            else:
                summary.noncompliantElementsCount += 1

    yield from responsible_summaries.values()


def cfg_element_statuses_storage_summaries(
    cfg_element_statuses: typing.Iterable[cmm.CfgElementStatusReport],
) -> typing.Generator[cmm.CfgStorageSummary, None, None]:

    storage_summaries = dict()

    def storage_summary(element_storage: str) -> cmm.CfgStorageSummary:
        if (summary := storage_summaries.get(element_storage)):
            return summary

        cfg_storage_summary = cmm.CfgStorageSummary(
            url=element_storage,
            noRuleAssigned=[],
            noStatus=[],
            assignedRuleRefersToUndefinedPolicy=[],
            noResponsibleAssigned=[],
            credentialsOutdated=[],
            credentialsNotOutdated=[],
            fullyCompliant=[],
        )
        storage_summaries[element_storage] = cfg_storage_summary
        return cfg_storage_summary

    for cfg_element_status in cfg_element_statuses:
        cfg_storage_summary = storage_summary(cfg_element_status.element_storage)
        evaluation_result = evaluate_cfg_element_status(cfg_element_status)

        if not evaluation_result.hasResponsible:
            cfg_storage_summary.noResponsibleAssigned.append(cfg_element_status)

        if not evaluation_result.hasRule:
            cfg_storage_summary.noRuleAssigned.append(cfg_element_status)

        elif evaluation_result.assignedRuleRefersToUndefinedPolicy:
            cfg_storage_summary.assignedRuleRefersToUndefinedPolicy.append(cfg_element_status)

        else:
            if evaluation_result.requiresStatus:
                if not evaluation_result.hasStatus:
                    cfg_storage_summary.noStatus.append(cfg_element_status)

                else:
                    if evaluation_result.credentialsOutdated:
                        cfg_storage_summary.credentialsOutdated.append(cfg_element_status)

                    else:
                        cfg_storage_summary.credentialsNotOutdated.append(cfg_element_status)

        if evaluation_result.fullyCompliant:
            cfg_storage_summary.fullyCompliant.append(cfg_element_status)
            cfg_storage_summary.compliantElementsCount += 1

        else:
            cfg_storage_summary.noncompliantElementsCount += 1

    yield from storage_summaries.values()


def create_report(
    cfg_element_statuses: typing.Iterable[cmm.CfgElementStatusReport],
):
    no_rule_assigned = []
    no_status = []
    assigned_rule_refers_to_undefined_policy = []
    no_responsible_assigned = []
    credentials_outdated = []
    credentials_not_outdated = []
    fully_compliant = []

    for cfg_element_status in cfg_element_statuses:
        evaluation_result = evaluate_cfg_element_status(cfg_element_status)

        if not evaluation_result.hasResponsible:
            no_responsible_assigned.append(cfg_element_status)

        if not evaluation_result.hasRule:
            no_rule_assigned.append(cfg_element_status)

        elif evaluation_result.assignedRuleRefersToUndefinedPolicy:
            assigned_rule_refers_to_undefined_policy.append(cfg_element_status)

        else:
            if evaluation_result.requiresStatus:
                if not evaluation_result.hasStatus:
                    no_status.append(cfg_element_status)

                else:
                    if evaluation_result.credentialsOutdated:
                        credentials_outdated.append(cfg_element_status)

                    else:
                        credentials_not_outdated.append(cfg_element_status)

        if evaluation_result.fullyCompliant:
            fully_compliant.append(cfg_element_status)

    def print_paragraph(header: str, statuses: typing.List[cmm.CfgElementStatusReport]):
        print(f'({len(statuses)}) {header}')
        print(2*'\n')

        for status in statuses:
            print(status.name)

        print('')
        print(40 * '-')

    print_paragraph(
        header='Cfg-Elements w/o assigned policy/rule',
        statuses=no_rule_assigned,
    )
    print_paragraph(
        header='Cfg-Elements w/o current status',
        statuses=no_status,
    )
    print_paragraph(
        header='Cfg-Elements with undefined policy',
        statuses=assigned_rule_refers_to_undefined_policy,
    )
    print_paragraph(
        header='Cfg-Elements w/o assigned responsible',
        statuses=no_responsible_assigned,
    )
    print_paragraph(
        header='Cfg-Elements w/ outdated credentials',
        statuses=credentials_outdated,
    )
    print_paragraph(
        header='Cfg-Elements with sufficiently recent credentials',
        statuses=credentials_not_outdated,
    )
    print_paragraph(
        header='Fully compliant cfg-elements *.*',
        statuses=fully_compliant,
    )


def cfg_compliance_status_to_es(
    es_client: ccc.elasticsearch.ElasticSearchClient,
    cfg_report_summary_gen: typing.Generator[cmm.CfgStorageSummary, None, None],
):
    for cfg_report_summary in cfg_report_summary_gen:
        cc_cfg_compliance_status = cfg_mgmt.metrics.CcCfgComplianceStatus.create(
            url=cfg_report_summary.url,
            compliant_count=cfg_report_summary.compliantElementsCount,
            non_compliant_count=cfg_report_summary.noncompliantElementsCount,
        )

        ccc.elasticsearch.metric_to_es(
            es_client=es_client,
            metric=cc_cfg_compliance_status,
            index_name=cfg_mgmt.metrics.index_name(cc_cfg_compliance_status),
        )


def cfg_compliance_storage_responsibles_to_es(
    es_client: ccc.elasticsearch.ElasticSearchClient,
    cfg_responsible_summary_gen: typing.Generator[cmm.CfgResponsibleSummary, None, None],
):
    for cfg_responsible_sum in cfg_responsible_summary_gen:
        cc_cfg_compliance_storage_responsibles = \
            cfg_mgmt.metrics.CcCfgComplianceStorageResponsibles.create(
            url=cfg_responsible_sum.url,
            compliant_count=cfg_responsible_sum.compliantElementsCount,
            non_compliant_count=cfg_responsible_sum.noncompliantElementsCount,
            responsible=cfg_responsible_sum.responsible,
        )

        ccc.elasticsearch.metric_to_es(
            es_client=es_client,
            metric=cc_cfg_compliance_storage_responsibles,
            index_name=cfg_mgmt.metrics.index_name(cc_cfg_compliance_storage_responsibles),
        )


def cfg_compliance_responsibles_to_es(
    es_client: ccc.elasticsearch.ElasticSearchClient,
    cfg_element_statuses: typing.Iterable[cmm.CfgElementStatusReport],
):
    for cfg_element_status in cfg_element_statuses:

        status_evaluation = evaluate_cfg_element_status(cfg_element_status)

        cc_cfg_compliance_responsible = cfg_mgmt.metrics.CcCfgComplianceResponsible.create(
            element_name=cfg_element_status.element_name,
            element_type=cfg_element_status.element_type,
            element_storage=cfg_element_status.element_storage,
            is_compliant=status_evaluation.fullyCompliant,
            responsible=cfg_element_status.responsible,
            rotation_method=cfg_element_status.policy.rotation_method,
            non_compliant_reasons=status_evaluation.nonCompliantReasons,
        )

        ccc.elasticsearch.metric_to_es(
            es_client=es_client,
            metric=cc_cfg_compliance_responsible,
            index_name=cfg_mgmt.metrics.index_name(cc_cfg_compliance_responsible),
        )


def iter_cfg_elements_requiring_rotation(
    cfg_elements: typing.Iterable[model.NamedModelElement],
    cfg_metadata: cmm.CfgMetadata,
    cfg_target: typing.Optional[cmm.CfgTarget]=None,
    element_filter: typing.Callable[[model.NamedModelElement], bool]=None,
    rotation_method: cmm.RotationMethod=None,
) -> typing.Generator[model.NamedModelElement, None, None]:
    for cfg_element in cfg_elements:
        if cfg_target and not cfg_target.matches(element=cfg_element):
            continue

        if element_filter and not element_filter(cfg_element):
            continue

        status = cmu.determine_status(
            element=cfg_element,
            policies=cfg_metadata.policies,
            rules=cfg_metadata.rules,
            responsibles=cfg_metadata.responsibles,
            statuses=cfg_metadata.statuses,
        )

        # hardcode rule: ignore elements w/o rule and policy
        if not status.policy or not status.rule:
            continue

        # hardcode: ignore all policies we cannot handle (currently, only MAX_AGE)
        if not status.policy.type is cmm.PolicyType.MAX_AGE:
            continue

        if rotation_method and status.policy.rotation_method is not rotation_method:
            continue

        # if there is no status, assume rotation be required
        if not status.status:
            yield cfg_element
            continue

        last_update = dp.isoparse(status.status.credential_update_timestamp)
        if status.policy.check(last_update=last_update, honour_grace_period=False):
            continue
        else:
            yield cfg_element


def generate_cfg_element_status_reports(
    cfg_dir: str,
    element_storage: str | None=None,
) -> list[cmm.CfgElementStatusReport]:
    '''
    If not passed explicitly, the element_storage defaults to cfg_dir.
    '''
    ci.util.existing_dir(cfg_dir)

    cfg_factory = model.ConfigFactory._from_cfg_dir(
        cfg_dir,
        disable_cfg_element_lookup=True,
    )

    cfg_metadata = cmm.cfg_metadata_from_cfg_dir(cfg_dir)

    policies = cfg_metadata.policies
    rules = cfg_metadata.rules
    statuses = cfg_metadata.statuses
    responsibles = cfg_metadata.responsibles

    if not element_storage:
        element_storage = cfg_dir

    return [
        cmu.determine_status(
            element=element,
            policies=policies,
            rules=rules,
            statuses=statuses,
            responsibles=responsibles,
            element_storage=element_storage,
        ) for element in cmu.iter_cfg_elements(cfg_factory=cfg_factory)
    ]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from concourse.model.job import (
    JobVariant,
)
from concourse.model.base import (
  AttributeSpec,
  Trait,
  TraitTransformer,
)


ATTRIBUTES = (
    AttributeSpec.optional(
        name='suppress_parallel_execution',
        default=None,
        doc='whether parallel executions of the same job should be prevented',
        type=bool,
    ),
)


class SchedulingTrait(Trait):
    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    # XXX: merge this with cron-trait
    def transformer(self):
        return SchedulingTraitTransformer()

    def suppress_parallel_execution(self):
        return self.raw.get('suppress_parallel_execution', None)


class SchedulingTraitTransformer(TraitTransformer):
    name = 'scheduling'

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # no-op
        pass




import cfg_mgmt.model as cmm
import github.compliance.issue as gci
import github.compliance.model as gcm


def scan_result_group_collection_for_outdated(results: tuple[gcm.CfgScanResult]) \
    -> gcm.ScanResultGroupCollection:
    issue_type = gci._label_outdated_credentials
    policy_violation = cmm.CfgElementPolicyViolation.CREDENTIALS_OUTDATED

    def classification_callback(result: gcm.CfgScanResult) -> gcm.Severity:
        return gcm.Severity.HIGH

    def findings_callback(result: gcm.CfgScanResult) -> bool:
        policy_violations = result.evaluation_result.nonCompliantReasons

        if policy_violation in policy_violations:
            return True

        return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=issue_type,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def scan_result_group_collection_for_no_status(results: tuple[gcm.CfgScanResult]) \
    -> gcm.ScanResultGroupCollection:
    issue_type = gci._label_no_status
    policy_violation = cmm.CfgElementPolicyViolation.NO_STATUS

    def classification_callback(result: gcm.CfgScanResult) -> gcm.Severity:
        return gcm.Severity.HIGH

    def findings_callback(result: gcm.CfgScanResult) -> bool:
        policy_violations = result.evaluation_result.nonCompliantReasons

        if policy_violation in policy_violations:
            return True

        return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=issue_type,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def scan_result_group_collection_for_no_responsible(results: tuple[gcm.CfgScanResult]) \
    -> gcm.ScanResultGroupCollection:
    issue_type = gci._label_no_responsible
    policy_violation = cmm.CfgElementPolicyViolation.NO_RESPONSIBLE

    def classification_callback(result: gcm.CfgScanResult) -> gcm.Severity:
        return gcm.Severity.MEDIUM

    def findings_callback(result: gcm.CfgScanResult) -> bool:
        policy_violations = result.evaluation_result.nonCompliantReasons

        if policy_violation in policy_violations:
            return True

        return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=issue_type,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def scan_result_group_collection_for_no_rule(results: tuple[gcm.CfgScanResult]) \
    -> gcm.ScanResultGroupCollection:
    issue_type = gci._label_no_rule
    policy_violation = cmm.CfgElementPolicyViolation.NO_RULE

    def classification_callback(result: gcm.CfgScanResult) -> gcm.Severity:
        return gcm.Severity.MEDIUM

    def findings_callback(result: gcm.CfgScanResult) -> bool:
        policy_violations = result.evaluation_result.nonCompliantReasons

        if policy_violation in policy_violations:
            return True

        return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=issue_type,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def scan_result_group_collection_for_undefined_policy(results: tuple[gcm.CfgScanResult]) \
    -> gcm.ScanResultGroupCollection:
    issue_type = gci._label_undefined_policy
    policy_violation = cmm.CfgElementPolicyViolation.ASSIGNED_RULE_REFERS_TO_UNDEFINED_POLICY

    def classification_callback(result: gcm.CfgScanResult) -> gcm.Severity:
        return gcm.Severity.MEDIUM

    def findings_callback(result: gcm.CfgScanResult) -> bool:
        policy_violations = result.evaluation_result.nonCompliantReasons

        if policy_violation in policy_violations:
            return True

        return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=issue_type,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )



import collections.abc
import dataclasses

import dacite

import oci.client as oc
import oci.model as om


def from_manifest(
    image_reference: om.OciImageReference,
    manifest: om.OciImageManifest,
    oci_client: oc.Client=None,
    base_platform: om.OciPlatform=None,
) -> om.OciPlatform:
    if base_platform:
        cfg = base_platform.as_dict()
    else:
        cfg = {}

    cfg |= oci_client.blob(
        image_reference=image_reference,
        digest=manifest.config.digest,
        stream=False, # we will need to json.load the (small) result anyhow
    ).json()

    return dacite.from_dict(
        data_class=om.OciPlatform,
        data=cfg,
    )


def from_single_image(
    image_reference: str | om.OciImageReference,
    oci_client: oc.Client=None,
    base_platform: om.OciPlatform=None,
) -> om.OciPlatform:
    '''
    determines the platform from a "single oci image" (i.e. an oci image which is _not_
    a multiarch image).
    '''
    image_reference = om.OciImageReference.to_image_ref(image_reference)

    manifest = oci_client.manifest(image_reference=image_reference)

    if not isinstance(manifest, om.OciImageManifest):
        raise ValueError(f'{image_reference=} did not yield OciImageManifest: {type(manifest)=}')

    return from_manifest(
        image_reference=image_reference,
        manifest=manifest,
        oci_client=oci_client,
        base_platform=base_platform,
    )


def single_platform_manifest(
    image_reference: om.OciImageReference | str,
    oci_client: oc.Client,
    platform: om.OciPlatform=None,
):
    '''
    returns a single-platform OCI Image Manifest for the given image_reference.
    lookup and validation depend on presence of platform argument.

    if given image-ref points to a single-arch manifest, the returned result will be identical
    to invoking `oci_client.manifest`. If platform argument is passed, and the discovered
    platform does not match, a `ValueError` will be raised.

    if given image-ref points to a multi-arch manifest, content-negotiation depends on presence of
    platform-argument. If absent, no preference will be stated (i.e. accept-header will not be set).
    Some Oci-Image-registries will return a single-arch manifest (thus saving a roundtrip).
    If platform is passed, preference for multi-arch will be stated via accept-header; the specified
    platform will be looked-up and returned. If not found, `ValueError` will be raised.
    '''
    image_reference = om.OciImageReference.to_image_ref(image_reference)

    if platform:
        accept = om.MimeTypes.prefer_multiarch
    else:
        accept = None

    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=accept,
    )

    if isinstance(manifest, om.OciImageManifest):
        if not platform:
            return manifest

        actual_platform = from_manifest(
            image_reference=image_reference,
            manifest=manifest,
            oci_client=oci_client,
        )

        if not actual_platform == platform:
            raise ValueError(f'{image_reference=} does not match {platform=}: {actual_platform=}')

        return manifest
    elif isinstance(manifest, om.OciImageManifestList):
        pass
    else:
        raise NotImplementedError(manifest)

    for manifest in manifest.manifests:
        manifest: om.OciImageManifestListEntry
        if manifest.platform == platform:
            break
    else:
        raise ValueError(f'{image_reference=} does not contain {platform=}')

    manifest_ref = f'{image_reference.ref_without_tag}@{manifest.digest}'
    return oci_client.manifest(image_reference=manifest_ref)


def iter_platforms(
    image_reference: str | om.OciImageReference,
    oci_client: oc.Client=None,
) -> collections.abc.Generator[tuple[om.OciImageReference, om.OciPlatform], None, None]:
    image_reference = om.OciImageReference.to_image_ref(image_reference)

    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=om.MimeTypes.prefer_multiarch,
    )

    if isinstance(manifest, om.OciImageManifest):
        platform = from_single_image(
            image_reference=image_reference,
            oci_client=oci_client,
        )
        yield (image_reference, platform)
        return
    elif isinstance(manifest, om.OciImageManifestList):
        manifest: om.OciImageManifestList
    else:
        raise NotImplementedError(type(manifest))

    prefix = image_reference.ref_without_tag

    for sub_manifest in manifest.manifests:
        platform_dict = dataclasses.asdict(sub_manifest)

        sub_manifest = oci_client.manifest(
            image_reference=(sub_img_ref := f'{prefix}@{sub_manifest.digest}'),
        )
        platform = from_single_image(
            image_reference=sub_img_ref,
            oci_client=oci_client,
        )

        # merge platform-dicts - the one from cfg-blob is assumed to be more specific
        platform_dict |= dataclasses.asdict(platform)

        platform = dacite.from_dict(
            data_class=om.OciPlatform,
            data=platform_dict,
        )

        yield (sub_img_ref, platform)



import collections.abc
import dataclasses
import enum
import hashlib
import json

import cnudie.util
import ocm
import gci.oci
import oci.client
import oci.model as om


class UploadMode(enum.StrEnum):
    SKIP = 'skip'
    FAIL = 'fail'
    OVERWRITE = 'overwrite'


def _oci_blob_ref_from_access(
    access: ocm.LocalBlobAccess,
    oci_client: oci.client.Client=None,
    oci_image_reference: om.OciImageReference=None,
) -> om.OciBlobRef:
    '''
    create OciBlobRef from LocalBlobAccess. If access has a globalAccess attribute, all
    values will be read from it and returned.
    Otherwise, the `size` attribute will be looked up using the (then required) oci_client. A
    blob with a digest matching the one from access.localReference (typically of form
    sha256:<hexdigest>) must be present at the given oci_image_reference.

    This function is useful for creating OCI Image Manifest layer entries from local blob access
    objects from Component Descriptors.
    '''
    if (global_access := access.globalAccess):
        return om.OciBlobRef(
            digest=global_access.digest,
            mediaType=global_access.mediaType,
            size=global_access.size,
        )
    if not oci_client or not oci_image_reference:
        raise ValueError(
            'oci_client and oci_image_reference must both be set if no globalAccess is present',
            access,
        )

    res = oci_client.head_blob(
        image_reference=oci_image_reference,
        digest=access.localReference,
        absent_ok=True,
    )

    if not res.ok:
        if res.status_code == 404:
            raise ValueError(
                f'{access.localReference=} not present at {oci_image_reference=}',
                access,
            )
        res.raise_for_status()

    length = int(res.headers.get('content-length'))

    return om.OciBlobRef(
        digest=access.localReference,
        mediaType=access.mediaType,
        size=length,
    )


def _iter_oci_blob_refs(
    component: ocm.Component,
    oci_client: oci.client.Client=None,
    oci_image_reference: om.OciImageReference=None,
) -> collections.abc.Generator[None, None, om.OciBlobRef]:
    for artefact in component.iter_artefacts():
        access = artefact.access
        if not isinstance(access, ocm.LocalBlobAccess):
            continue

        blob_ref = _oci_blob_ref_from_access(
            access=access,
            oci_client=oci_client,
            oci_image_reference=oci_image_reference,
        )
        yield blob_ref


def upload_component_descriptor(
    component_descriptor: ocm.ComponentDescriptor | ocm.Component,
    on_exist: UploadMode | str=UploadMode.SKIP,
    ocm_repository: ocm.OciOcmRepository | str=None,
    oci_client: oci.client.Client=None,
):
    if not oci_client:
        try:
            import ccc.oci
        except ImportError:
            raise ValueError('must pass-in oci_client')
        oci_client = ccc.oci.oci_client()

    on_exist = UploadMode(on_exist)

    if isinstance(component_descriptor, ocm.Component):
        component_descriptor = ocm.ComponentDescriptor(
            component=component_descriptor,
            meta=ocm.Metadata(),
            signatures=[],
        )

    component = component_descriptor.component

    schema_version = component_descriptor.meta.schemaVersion
    if isinstance(schema_version, enum.Enum):
        # hack: convert from gci.componentmodel
        schema_version = ocm.SchemaVersion(schema_version.value)
    else:
        schema_version = ocm.SchemaVersion(schema_version)

    if not schema_version is ocm.SchemaVersion.V2:
        raise RuntimeError(f'unsupported component-descriptor-version: {schema_version=}')

    if ocm_repository:
        if isinstance(ocm_repository, str):
            ocm_repository = ocm.OciOcmRepository(baseUrl=ocm_repository)
        elif isinstance(ocm_repository, ocm.OciOcmRepository):
            pass
        else:
            raise TypeError(type(ocm_repository))

        if not component.current_ocm_repo == ocm_repository:
            component.repositoryContexts.append(ocm_repository)

    target_ref = cnudie.util.oci_artefact_reference(component)

    if on_exist in (UploadMode.SKIP, UploadMode.FAIL):
        # check whether manifest exists (head_manifest does not return None)
        if oci_client.head_manifest(image_reference=target_ref, absent_ok=True):
            if on_exist is UploadMode.SKIP:
                return
            if on_exist is UploadMode.FAIL:
                # XXX: we might still ignore it, if the to-be-uploaded CD is equal to the existing
                # one
                raise ValueError(f'{target_ref=} already existed')
    elif on_exist is UploadMode.OVERWRITE:
        pass
    else:
        raise NotImplementedError(on_exist)

    raw_fobj = gci.oci.component_descriptor_to_tarfileobj(component_descriptor)
    cd_digest = hashlib.sha256()
    while (chunk := raw_fobj.read(4096)):
        cd_digest.update(chunk)

    cd_octets = raw_fobj.tell()
    cd_digest = cd_digest.hexdigest()
    cd_digest_with_alg = f'sha256:{cd_digest}'
    raw_fobj.seek(0)

    oci_client.put_blob(
        image_reference=target_ref,
        digest=cd_digest_with_alg,
        octets_count=cd_octets,
        data=raw_fobj,
    )

    cfg = gci.oci.ComponentDescriptorOciCfg(
        componentDescriptorLayer=gci.oci.ComponentDescriptorOciBlobRef(
            digest=cd_digest_with_alg,
            size=cd_octets,
        )
    )
    cfg_raw = json.dumps(dataclasses.asdict(cfg)).encode('utf-8')
    cfg_octets = len(cfg_raw)
    cfg_digest = hashlib.sha256(cfg_raw).hexdigest()
    cfg_digest_with_alg = f'sha256:{cfg_digest}'

    oci_client.put_blob(
        image_reference=target_ref,
        digest=cfg_digest_with_alg,
        octets_count=cfg_octets,
        data=cfg_raw,
    )

    local_blob_layers = { # use set for deduplication
        blob_ref for blob_ref in
        _iter_oci_blob_refs(
            component=component,
            oci_client=oci_client,
            oci_image_reference=target_ref,
        )
    }

    manifest = om.OciImageManifest(
        config=gci.oci.ComponentDescriptorOciCfgBlobRef(
            digest=cfg_digest_with_alg,
            size=cfg_octets,
        ),
        layers=[
            gci.oci.ComponentDescriptorOciBlobRef(
                digest=cd_digest_with_alg,
                size=cd_octets,
            ),
        ] + list(local_blob_layers),
    )

    manifest_dict = manifest.as_dict()
    manifest_bytes = json.dumps(manifest_dict).encode('utf-8')

    oci_client.put_manifest(
        image_reference=target_ref,
        manifest=manifest_bytes,
    )

    return manifest_bytes



import datetime
import logging
import tarfile
import typing

import ocm

import ci.log
import ci.util
import cnudie.iter
import cnudie.retrieve
import delivery.client
import delivery.util
import dso.model as dm
import github.compliance.issue as gciss
import github.compliance.model as gcm
import oci.client
import oci.model
import tarutil
import unixutil.model as um
import unixutil.scan as us

logger = logging.getLogger('os-id')
ci.log.configure_default_logging()


def determine_os_ids(
    component_descriptor: ocm.ComponentDescriptor,
    oci_client: oci.client.Client,
    lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    delivery_service_client: delivery.client.DeliveryServiceClient,
) -> typing.Generator[gcm.OsIdScanResult, None, None]:
    for component, resource in cnudie.iter.iter_resources(
        component=component_descriptor,
        lookup=lookup,
    ):
        if resource.type != ocm.ArtefactType.OCI_IMAGE:
            continue

        if not resource.access:
            continue

        if resource.access.type != ocm.AccessType.OCI_REGISTRY:
            continue

        yield base_image_os_id(
            oci_client=oci_client,
            delivery_service_client=delivery_service_client,
            component=component,
            resource=resource,
        )


def base_image_os_id(
    oci_client: oci.client.Client,
    delivery_service_client: delivery.client.DeliveryServiceClient,
    component: ocm.Component,
    resource: ocm.Resource,
) -> gcm.OsIdScanResult:
    # shortcut scan if there is already a scan-result
    if delivery_service_client:
        scan_result = delivery_service_client.metadata(
            component=component,
            artefact=resource,
            types=('os_ids',),
        )
        scan_result = tuple(scan_result)
        if len(scan_result) > 1:
            logger.warning(
                f'found more than one scanresult for {component.name}:{resource.name}'
            )
        if scan_result:
            scan_result = scan_result[0]

            if not isinstance(scan_result.data, dm.OsID):
                logger.warning('unexpected scan-result-type: {scan_result=}')
            else:
                scan_result.data: dm.OsID

                return gcm.OsIdScanResult(
                    scanned_element=cnudie.iter.ResourceNode(
                        path=(cnudie.iter.NodePathEntry(component),),
                        resource=resource,
                    ),
                    os_id=scan_result.data.os_info,
                    skip_upload_to_deliverydb=True, # no need to re-upload (got it from there)
                )

    # there was no scanresult, so we have to scan
    image_reference = resource.access.imageReference

    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=oci.model.MimeTypes.prefer_multiarch,
    )

    # if multi-arch, randomly choose first entry (assumption: all variants have same os/version)
    if isinstance(manifest, oci.model.OciImageManifestList):
        manifest: oci.model.OciImageManifestList
        manifest: oci.model.OciBlobRef = manifest.manifests[0]
        image_reference = oci.model.OciImageReference(image_reference)
        manifest = oci_client.manifest(image_reference.ref_without_tag + '@' + manifest.digest)

    last_os_info = None

    for layer in manifest.layers:
        layer_blob = oci_client.blob(
            image_reference=image_reference,
            digest=layer.digest,
        )
        fileproxy = tarutil.FilelikeProxy(
            layer_blob.iter_content(chunk_size=tarfile.BLOCKSIZE)
        )
        tf = tarfile.open(fileobj=fileproxy, mode='r|*')
        if (os_info := us.determine_osinfo(tf)):
            last_os_info = os_info

    if not last_os_info:
        # if we could not determine os-info, upload a dummy os-info (with all entries set to None)
        # to keep track of the failed scan attempt
        last_os_info = um.OperatingSystemId()

    # pylint: disable=E1123
    return gcm.OsIdScanResult(
        scanned_element=cnudie.iter.ResourceNode(
            path=(cnudie.iter.NodePathEntry(component),),
            resource=resource,
        ),
        os_id=last_os_info,
    )


def scan_result_group_collection_for_outdated_os_ids(
    results: tuple[gcm.OsIdScanResult],
    delivery_svc_client: delivery.client.DeliveryServiceClient,
):
    os_names = {r.os_id.ID for r in results if r.os_id.ID}
    os_infos = {
        os_name: info
        for os_name in os_names
        if (info := delivery_svc_client.os_release_infos(os_id=os_name, absent_ok=True)) is not None
    }

    def classification_callback(result: gcm.OsIdScanResult):
        os_id = result.os_id
        if not os_id.ID in os_infos:
            return None

        if os_id.is_distroless: # hardcode: never require distroless-images to be updated
            return None

        if delivery.util.branch_reached_eol(
            os_id=os_id,
            os_infos=os_infos[os_id.ID],
        ):
            return gcm.Severity.HIGH
        elif delivery.util.update_available(
            os_id=os_id,
            os_infos=os_infos[os_id.ID],
            ignore_if_patchlevel_is_next_to_greatest=True,
        ):
            return gcm.Severity.MEDIUM
        else:
            return None

    def findings_callback(result: gcm.OsIdScanResult):
        os_id = result.os_id
        if not os_id.ID in os_infos:
            return None

        if os_id.is_distroless: # hardcode: never require distroless-images to be updated
            return None

        relation = result.scanned_element.resource.relation

        if delivery.util.branch_reached_eol(
            os_id=os_id,
            os_infos=os_infos[os_id.ID],
        ):
            return True

        if not relation is ocm.ResourceRelation.LOCAL:
            logger.info(f'{result.scanned_element.resource.name=} '
                f'is not "local" - will ignore findings')
            return False

        if delivery.util.update_available(
            os_id=os_id,
            os_infos=os_infos[os_id.ID],
            ignore_if_patchlevel_is_next_to_greatest=True,
        ):
            return True
        else:
            return False

    return gcm.ScanResultGroupCollection(
        results=tuple(results),
        issue_type=gciss._label_os_outdated,
        classification_callback=classification_callback,
        findings_callback=findings_callback,
    )


def upload_to_delivery_db(
    db_client: delivery.client.DeliveryServiceClient,
    resource: ocm.Resource,
    component: ocm.Component,
    os_info: um.OperatingSystemId,
):
    artefact_ref = dm.component_artefact_id_from_ocm(
        component=component,
        artefact=resource,
    )

    now = datetime.datetime.now()
    meta = dm.Metadata(
        datasource=dm.Datasource.CC_UTILS,
        type=dm.Datatype.OS_IDS,
        creation_date=now,
        last_update=now,
    )

    os_id = dm.OsID(
        os_info=os_info,
    )
    artefact_metadata = dm.ArtefactMetadata(
        artefact=artefact_ref,
        meta=meta,
        data=os_id,
    )

    db_client.update_metadata(data=[artefact_metadata])



import collections
import dataclasses
import datetime
import enum
import functools
import typing

import github3.repos

import cfg_mgmt.model as cmm
import cnudie.iter
import delivery.client
import ocm
import github.compliance.milestone as gcmi
import unixutil.model


class Severity(enum.IntEnum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 4
    CRITICAL = 8
    BLOCKER = 16

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return "'" + self.name.lower() + "'"

    @staticmethod
    def parse(value: int | str):
        '''
        convenience method that will behave like default-c'tor, but also accept str-values
        (ignoring type for additional convenience)
        '''
        if isinstance(value, str):
            return Severity[value.upper()]

        return Severity(value)


@dataclasses.dataclass(frozen=True)
class MaxProcessingTimesDays:
    '''
    defines maximum processing time in days, based on issue "criticality"

    used for deserialisation from pipeline-definitions
    '''
    blocker: int = 0
    very_high_or_greater: int = 30
    high: int = 30
    medium: int = 90
    low: int = 120

    def for_severity(self, severity: Severity):
        if severity is Severity.BLOCKER:
            return self.blocker
        elif severity is Severity.CRITICAL:
            return self.very_high_or_greater
        elif severity is Severity.HIGH:
            return self.high
        elif severity is Severity.MEDIUM:
            return self.medium
        elif severity is Severity.LOW:
            return self.low


class ScanState(enum.Enum):
    '''
    indicates the scan outcome of a scan (regardless of yielded contents).

    SUCCEEDED:  scan succeeded without errors (but potentially with findings)
    SKIPPED:    No scan took place (we have an earlier scan that has not been invalidated yet)
    FAILED:     scan failed (which typically implies there are not scan results)
    '''
    SUCCEEDED = 'succeeded'
    SKIPPED = 'skipped'
    FAILED = 'failed'


Target = cnudie.iter.ResourceNode | cnudie.iter.SourceNode | cmm.CfgElementStatusReport


@dataclasses.dataclass(kw_only=True)
class ScanResult:
    scanned_element: Target
    discovery_date: datetime.date = datetime.date.today()

    state: ScanState = ScanState.SUCCEEDED

    @property
    def scan_succeeded(self) -> bool:
        return self.state in [ScanState.SUCCEEDED, ScanState.SKIPPED]

    @property
    def severity(self) -> Severity:
        # has to be defined by enclosing class
        pass

    def calculate_latest_processing_date(
        self,
        max_processing_days: MaxProcessingTimesDays=None,
        delivery_svc_client: delivery.client.DeliveryServiceClient=None,
        repository: github3.repos.Repository=None,
    ) -> datetime.date | None:
        # explicitly check for `None` in case severity is `0`
        if self.severity is None:
            return None

        if not max_processing_days:
            max_processing_days = MaxProcessingTimesDays()
        max_days = max_processing_days.for_severity(severity=self.severity)
        date = self.discovery_date + datetime.timedelta(days=max_days)

        if delivery_svc_client and repository:
            try:
                target_sprints = gcmi.target_sprints(
                    delivery_svc_client=delivery_svc_client,
                    latest_processing_date=date,
                    sprints_count=2,
                )
                target_milestone, _ = gcmi.find_or_create_sprint_milestone(
                    repo=repository,
                    sprints=target_sprints,
                )
                if target_milestone:
                    date = target_milestone.due_on.date()
            except Exception:
                import traceback
                traceback.print_exc()

        return date


@dataclasses.dataclass
class OsIdScanResult(ScanResult):
    os_id: unixutil.model.OperatingSystemId
    skip_upload_to_deliverydb: bool = False


@dataclasses.dataclass
class CfgScanResult(ScanResult):
    evaluation_result: cmm.CfgStatusEvaluationResult


FindingsCallback = typing.Callable[[ScanResult], bool]
'''
callback type accepting a ScanResult; expected to return True iff argument has a "finding" and False
otherwise.

Definition of "finding" is type-specific
'''
ClassificationCallback = typing.Callable[[ScanResult], Severity]


def is_ocm_artefact_node(
    element: cnudie.iter.SourceNode | cnudie.iter.ResourceNode | object,
):
    if isinstance(element, (cnudie.iter.SourceNode, cnudie.iter.ResourceNode)):
        return True

    return False


@dataclasses.dataclass
class ScanResultGroup:
    '''
    a group of scan results (grouped by scanned_element + latest processing date if it exists)
    grouping is done so alike scanned_elements are grouped into common "reporting
    targets" (github issues if used in the context of this package)

    components and artifacts are understood as defined by the OCM (ocm)

    ScanResultGroup caches calculated values to reduce amount of (potentially expensive) callbacks.
    Altering `results`, or external state passed-in callbacks rely on will thus result in
    inconsistent state.
    '''
    name: str
    # {component.name}:{component.version}:{artifact.name}:{artifact.version} |
    # {component.name}:{component.version}:{artifact.name}:{artifact.version}:{processing_date} |
    # {cfg_element_storage}/{cfg_element_type}/{cfg_element_name}
    results: tuple[ScanResult]
    issue_type: str
    findings_callback: FindingsCallback
    classification_callback: ClassificationCallback

    @property
    def results_with_successful_scans(self):
        return tuple(r for r in self.results if r.scan_succeeded)

    @property
    def component(self) -> ocm.Component:
        result = self.results[0]
        if not is_ocm_artefact_node(result.scanned_element):
            raise RuntimeError('property not allowed to be used if scanned_element is '
                'not either a ResourceNode or a SourceNode')

        return result.scanned_element.component

    @property
    def artifact(self) -> ocm.Source | ocm.Resource:
        result = self.results[0]
        if not is_ocm_artefact_node(result.scanned_element):
            raise RuntimeError('property not allowed to be used if scanned_element is '
                'not either a ResourceNode or a SourceNode')

        return artifact_from_node(result.scanned_element)

    @functools.cached_property
    def has_findings(self) -> bool:
        for r in self.results_with_successful_scans:
            if r.state is ScanState.SKIPPED:
                continue
            if self.findings_callback(r):
                return True
        else:
            return False

    @functools.cached_property
    def has_scan_errors(self) -> bool:
        for result in self.results:
            if not result.scan_succeeded:
                return True

        return False

    @functools.cached_property
    def has_attempted_scans(self) -> bool:
        for result in self.results:
            if not result.state is ScanState.SKIPPED:
                return True
        return False

    @functools.cached_property
    def worst_severity(self) -> Severity:
        if not self.has_findings:
            return None
        classifications = [self.classification_callback(r) for r in self.results_with_findings]
        return max(classifications)

    @functools.cached_property
    def worst_result(self) -> ScanResult:
        if not self.has_findings:
            return None

        worst_severity = self.worst_severity

        for result in self.results_with_findings:
            if self.classification_callback(result) is worst_severity:
                return result

        return None

    @functools.cached_property
    def results_with_findings(self) -> tuple[ScanResult]:
        return tuple(
            r for r in self.results_with_successful_scans
            if r.state is not ScanState.SKIPPED and self.findings_callback(r)
        )

    @functools.cached_property
    def results_without_findings(self) -> tuple[ScanResult]:
        return tuple((
            r for r in self.results_with_successful_scans
            if r.state is not ScanState.SKIPPED and not self.findings_callback(r)
        ))


@dataclasses.dataclass
class ScanResultGroupCollection:
    results: tuple[ScanResult]
    issue_type: str
    classification_callback: ClassificationCallback
    findings_callback: FindingsCallback
    max_processing_days: MaxProcessingTimesDays = None
    delivery_svc_client: delivery.client.DeliveryServiceClient = None
    repository: github3.repos.Repository = None

    @property
    def result_groups(self) -> tuple[ScanResultGroup]:
        if not self.results:
            return ()

        grouped_results = collections.defaultdict(list)

        import github.compliance.issue as gci
        for result in self.results:
            if is_ocm_artefact_node(result.scanned_element):
                c = result.scanned_element.component
                a = artifact_from_node(result.scanned_element)
                if (self.issue_type == gci._label_bdba or
                    self.issue_type == gci._label_licenses):
                    group_name = f'{c.name}:{c.version}:{a.name}:{a.version}'
                    if (latest_processing_date := result.calculate_latest_processing_date(
                        max_processing_days=self.max_processing_days,
                        delivery_svc_client=self.delivery_svc_client,
                        repository=self.repository,
                    )):
                        group_name += f':{latest_processing_date}'
                else:
                    group_name = f'{c.name}:{a.name}'
            elif isinstance(result.scanned_element, cmm.CfgElementStatusReport):
                group_name = result.scanned_element.name
            else:
                raise TypeError(result)

            grouped_results[group_name].append(result)

        return tuple((
            ScanResultGroup(
                name=group_name,
                results=results,
                issue_type=self.issue_type,
                findings_callback=self.findings_callback,
                classification_callback=self.classification_callback,
            ) for group_name, results in grouped_results.items()
        ))

    @property
    def result_groups_with_findings(self) -> tuple[ScanResultGroup]:
        return tuple(
            (rg for rg in self.result_groups if rg.has_findings)
        )

    @property
    def result_groups_without_findings(self) -> tuple[ScanResultGroup]:
        return tuple(
            (
                rg for rg in self.result_groups
                if not rg.has_findings and not rg.has_scan_errors
            )
        )

    @property
    def result_groups_with_scan_errors(self) -> tuple[ScanResultGroup]:
        return tuple(
            (rg for rg in self.result_groups if rg.has_scan_errors)
        )


def artifact_from_node(
    node: cnudie.iter.ResourceNode | cnudie.iter.SourceNode,
) -> ocm.Source | ocm.Resource:
    if isinstance(node, cnudie.iter.SourceNode):
        return node.source
    elif isinstance(node, cnudie.iter.ResourceNode):
        return node.resource
    else:
        raise TypeError(node)



'''
model classes and utils for dealing w/ legacy oci-images from legacy-version of docker
'''

import dataclasses
import datetime
import hashlib
import random

dc = dataclasses.dataclass


@dc
class DockerContainerCfg:
  Image: str # sha256-hash

  ArgsEscaped: bool = False
  AttachStderr: bool = False
  AttachStdin: bool = False
  AttachStdout: bool = False
  Cmd: list[str] | str | None = None
  Entrypoint: list[str] | str | None = None
  Domainname: str = ''
  Env: list[str] | None = (
      'PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin',
  )
  Hostname: str = ''
  Labels: dict[str, str] | None = dataclasses.field(default_factory=dict)
  OnBuild: list[str] | None = None
  OpenStdin: bool = False
  StdinOnce: bool = False
  Tty: bool = False
  User: str = ''
  Volumes: list[str] | None = None
  WorkingDir: str = ''


@dc
class Docker_Fs:
    diff_ids: list[str] = dataclasses.field(default_factory=list) # [layer-sha256-digests]
    type: str = 'layers'


@dc
class DockerCfg:
  '''
  a cfg as created / understood by docker (use as cfg in oci-manifest)
  '''
  config: DockerContainerCfg
  container: str # container-hash
  container_config: DockerContainerCfg
  created: str # iso8601-ts
  rootfs: Docker_Fs | None

  architecture: str = 'amd64'
  docker_version: str = '18.09.7'
  history: tuple[dict] = ()
  os: str = 'linux'


def docker_cfg():
  now_ts = datetime.datetime.now().isoformat() + 'Z'
  container_id = hashlib.sha256(f'{random.randint(0, 2 ** 32)}'.encode('utf-8')).hexdigest()

  cfg = DockerContainerCfg(Image=f'sha256:{container_id}')

  return DockerCfg(
    config=cfg,
    container=container_id,
    container_config=cfg,
    created=now_ts,
    rootfs=Docker_Fs(), # TODO: check whether we need to pass-in layer-hash-digests
  )



'''
utils for management of configuration provisioned in CICD-Clusters
(cfg and secrets stored and maintained in secret/private "cfg-repositories", replicated to
"secrets-server")
'''

import typing

revert_function = typing.Callable[[], None]



'''
collection of workarounds required to deal w/ different OCI Registries' idiosyncrazies
'''

import dataclasses
import hashlib
import json

import oci.client as oc
import oci.model as om


def _cfg_blob_non_empty_history_layers(cfg_blob: dict) -> list[dict]:
    history = cfg_blob['history']

    non_empty = [
        entry for entry in history if not entry.get('empty_layer', False)
    ]

    return non_empty


def is_cfg_blob_sane(
    manifest: om.OciImageManifest,
    cfg_blob: bytes | dict,
) -> bool:
    if isinstance(cfg_blob, bytes) or isinstance(cfg_blob, str):
        cfg_blob = json.loads(cfg_blob)
    if not isinstance(cfg_blob, dict):
        raise ValueError(cfg_blob)

    non_empty_layers_in_cfg_blob = len(_cfg_blob_non_empty_history_layers(cfg_blob=cfg_blob))

    if non_empty_layers_in_cfg_blob < len(manifest.layers):
        return False

    return True


def sanitise_cfg_blob(
    manifest: om.OciImageManifest,
    cfg_blob: bytes | dict,
) -> bytes | dict:
    '''
    returns a sanitised form of the passed cfg-blob. If the passed cfg-blob was already considered
    "sane", the returned object is identical to the passed-in cfg_blob argument.

    otherwise, a sanitised cfg-blob is returned as a utf-8-encoded `bytes` object.
    '''
    if is_cfg_blob_sane(manifest=manifest, cfg_blob=cfg_blob):
        return cfg_blob

    if isinstance(cfg_blob, bytes) or isinstance(cfg_blob, str):
        cfg_blob = json.loads(cfg_blob)

    cfg_blob_nonempty_layers = _cfg_blob_non_empty_history_layers(cfg_blob=cfg_blob)
    missing_history_entries = len(manifest.layers) - len(cfg_blob_nonempty_layers)

    if not cfg_blob_nonempty_layers:
        raise ValueError('cannot duplicate fake history-entries w/o at least one non-empty entry')

    # arbitrarily choose first entry to duplicate
    history_entry = cfg_blob_nonempty_layers[0]

    for _ in range(missing_history_entries):
        cfg_blob['history'].append(history_entry)

    return json.dumps(cfg_blob).encode('utf-8')


def sanitise_image(
    image_ref: str | om.OciImageReference,
    oci_client: oc.Client,
):
    manifest = oci_client.manifest(image_reference=image_ref)
    cfg_blob = oci_client.blob(image_reference=image_ref, digest=manifest.config.digest).content

    if is_cfg_blob_sane(manifest=manifest, cfg_blob=cfg_blob):
        return image_ref

    sanitised_cfg_blob = sanitise_cfg_blob(manifest=manifest, cfg_blob=cfg_blob)
    cfg_blob_digest = 'sha256:' + hashlib.sha256(sanitised_cfg_blob).hexdigest()

    oci_client.put_blob(
        image_ref,
        digest=cfg_blob_digest,
        octets_count=len(sanitised_cfg_blob),
        data=sanitised_cfg_blob,
    )

    manifest = dataclasses.replace(
        manifest,
        config=dataclasses.replace(
            manifest.config,
            digest=cfg_blob_digest,
            size=len(sanitised_cfg_blob),
        ),
    )

    manifest_bytes = json.dumps(manifest.as_dict()).encode('utf-8')

    oci_client.put_manifest(image_reference=image_ref, manifest=manifest_bytes)

    manifest_dig = 'sha256:' + hashlib.sha256(manifest_bytes).hexdigest()
    img_ref: om.OciImageReference = om.OciImageReference.to_image_ref(image_ref)

    patched_img_ref = f'{img_ref.ref_without_tag}@{manifest_dig}'

    return patched_img_ref



import copy
import dataclasses
import dateutil.parser
import typing
import logging

import msal
import dacite
import requests

import cfg_mgmt
import ci.log
import model
import model.azure

from cfg_mgmt.model import CfgQueueEntry


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


@dataclasses.dataclass
class PasswordCredential:
    displayName: str
    keyId: str
    endDateTime: str
    startDateTime: typing.Optional[str]
    customKeyIdentifier: typing.Optional[str]
    hint: typing.Optional[str]
    secretText: typing.Optional[str]  # the actual password, if present

    def end_DateTime(self):
        return dateutil.parser.isoparse(self.endDateTime)

    def start_DateTime(self):
        return dateutil.parser.isoparse(self.startDateTime)


def _get_access_token_for_principal(
    service_principal: model.azure.AzureServicePrincipal,
):

    app = msal.ConfidentialClientApplication(
        client_id=service_principal.client_id(),
        authority=f'https://login.microsoftonline.com/{service_principal.tenant_id()}',
        client_credential=service_principal.client_secret(),
    )
    scopes = ['https://graph.microsoft.com/.default',]
    response = app.acquire_token_for_client(scopes=scopes)

    if 'access_token' in response:
        access_token = response['access_token']
    elif 'error' in response:
        raise RuntimeError(
            f'Error authenticating with Microsoft Graph: {response.get("error_description")}'
        )
    else:
        raise RuntimeError(f'Unexpected response: {response}')

    return access_token


def _retrieve_password_credentials(
    service_principal: model.azure.AzureServicePrincipal,
    access_token: str,
    azure_app_root_url: str = 'https://graph.microsoft.com/v1.0/applications'
) -> typing.Iterable[PasswordCredential]:
    response = requests.get(
        url=f'{azure_app_root_url}/{service_principal.object_id()}',
        headers={'Authorization': f'Bearer {access_token}'},
        timeout=(4, 31),
    )
    response_dict = response.json()
    if not (password_credentials := response_dict.get('passwordCredentials')):
        # no keys
        logger.warning(
            f"No password-credentials found for service principal '{service_principal.name()}'"
        )
        return []

    return [PasswordCredential(**e) for e in password_credentials]


def _add_password_credential(
    service_principal: model.azure.AzureServicePrincipal,
    display_name: str,
    access_token: str,
    azure_app_root_url: str = 'https://graph.microsoft.com/v1.0/applications'
) -> PasswordCredential:
    '''Add a new password credential to the given service principal and return the corresponding
    PasswordCredential object

    Note: This will be the only time that the generated `secretText` (i.e. the actual password) is
    available - no other calls to the API populate this field.
    '''
    body = {"passwordCredential": {"displayName": display_name}}

    response = requests.post(
        f'{azure_app_root_url}/{service_principal.object_id()}/addPassword',
        json=body,
        headers={'Authorization': f'Bearer {access_token}'},
        timeout=(4, 31),
    )
    response.raise_for_status()
    response_dict = response.json()

    return dacite.from_dict(
        data_class=PasswordCredential,
        data=response_dict,
    )


def _remove_password_credential(
    service_principal: model.azure.AzureServicePrincipal,
    key_id: str,
    access_token: str,
    azure_app_root_url: str = 'https://graph.microsoft.com/v1.0/applications'
):
    response = requests.post(
        f'{azure_app_root_url}/{service_principal.object_id()}/removePassword',
        json={"keyId": key_id},
        headers={'Authorization': f'Bearer {access_token}'},
        timeout=(4, 31),
    )
    # successful request returns "204 No content"
    response.raise_for_status()


def rotate_cfg_element(
    cfg_element: model.azure.AzureServicePrincipal,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:

    access_token = _get_access_token_for_principal(
        service_principal=cfg_element,
    )

    known_password_credentials = _retrieve_password_credentials(
        service_principal=cfg_element,
        access_token=access_token,
    )

    matching_password_credentials = [
        p for p in known_password_credentials if cfg_element.client_secret().startswith(p.hint)
    ]

    # since we used the client secret to authenticate against the API we can be certain that
    # at least one password matches
    if len(matching_password_credentials) > 1:
        raise RuntimeError('Unable to unambiguously determine old password credential, aborting.')

    current_password_credential = matching_password_credentials[0]
    new_password_credential = _add_password_credential(
        service_principal=cfg_element,
        display_name='CI Access',
        access_token=access_token,
    )

    raw_cfg = copy.deepcopy(cfg_element.raw)
    new_element = model.azure.AzureServicePrincipal(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )
    new_element.raw['client_secret'] = new_password_credential.secretText

    secret_id = {'keyId': current_password_credential.keyId}

    def revert_function():
        _remove_password_credential(
            service_principal=cfg_element,
            key_id=new_password_credential.keyId,
            access_token=access_token,  # token is good for 20 minutes by default,
                                        # so it's safe to reuse here
        )

    return revert_function, secret_id, new_element


def delete_config_secret(
    cfg_element: model.azure.AzureServicePrincipal,
    cfg_factory: model.ConfigFactory,
    cfg_queue_entry: CfgQueueEntry,
) -> model.azure.AzureServicePrincipal | None:
    key_id = cfg_queue_entry.secretId['keyId']
    access_token = _get_access_token_for_principal(service_principal=cfg_element)
    _remove_password_credential(
        service_principal=cfg_element,
        key_id=key_id,
        access_token=access_token,
    )
    return None




import logging
import traceback

import concourse.client.api
import concourse.client.model
import concourse.enumerator
import concourse.replicator


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def trigger_resource_check(
    concourse_api: concourse.client.api.ConcourseApiBase,
    resources,
):
    logger.debug('trigger_resource_check')
    for resource in resources:
        logger.info('triggering resource check for: ' + resource.name)
        try:
            concourse_api.trigger_resource_check(
                pipeline_name=resource.pipeline_name(),
                resource_name=resource.name,
            )
        except Exception:
            traceback.print_exc()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from ci.util import not_none
from concourse.model.base import (
    AttributeSpec,
    Trait,
    TraitTransformer
)
from concourse.model.job import (
    JobVariant,
)


ATTRIBUTES = (
    AttributeSpec.optional(
        name='build_logs_to_retain',
        default=1000,
        doc='the amount of build logs to retain before log rotation occurs',
        type=int,
    ),
    AttributeSpec.optional(
        name='public_build_logs',
        default=False,
        doc='whether or not build logs are accessible to unauthenticated users',
        type=bool,
    ),
)


class OptionsTrait(Trait):
    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def build_logs_to_retain(self):
        return self.raw['build_logs_to_retain']

    def public_build_logs(self):
        return self.raw['public_build_logs']

    def transformer(self):
        return OptionsTraitTransformer(trait=self)


class OptionsTraitTransformer(TraitTransformer):
    name = 'options'

    def __init__(self, trait: OptionsTrait, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trait = not_none(trait)

    def process_pipeline_args(self, pipeline_args: JobVariant):
        pass



import collections.abc
import contextlib
import io
import queue
import threading


def normalise_image_reference(image_reference: str):
    if not isinstance(image_reference, str):
        raise ValueError(image_reference)

    if image_reference.startswith('/'):
        # keep relative reference as is
        return image_reference

    parts = image_reference.split('/')

    left_part = parts[0]
    # heuristically check if we have a (potentially) valid hostname
    if '.' not in left_part.split(':')[0]:
        # insert 'library' if only image name was given
        if len(parts) == 1:
            parts.insert(0, 'library')

        # probably, the first part is not a hostname; inject default registry host
        parts.insert(0, 'registry-1.docker.io')

    # of course, docker.io gets special handling
    if parts[0] == 'docker.io':
        parts[0] = 'registry-1.docker.io'

    return '/'.join(parts)


def urljoin(*parts):
    if len(parts) == 1:
        return parts[0]
    first = parts[0]
    last = parts[-1]
    middle = parts[1:-1]

    first = first.rstrip('/')
    middle = list(map(lambda s: s.strip('/'), middle))
    last = last.lstrip('/')

    return '/'.join([first] + middle + [last])


class _TeeFilelikeProxy:
    '''
    Takes a filelike object (which may be a non-seekable stream) and patches its
    read-method in such a way that upon every read, a copy of the read data is emitted
    through a generator.

    The generator will signal to be exhausted after the passed filelike object has been
    read until EOF.

    Note that the generator *MUST* be consumed in a different thread. Also note that if the
    generator is _not_ consumed, every stream read chunk will permanently be stored in
    this object's queue.

    In the context of the oci package, this is particularly useful for the special-case
    of converting container images w/ a legacy "v1"-manifest to the v2-version, as we need
    to both re-upload the gzip-compressed layer-blobs unmodified, but also calculate digest-hashes
    for uncompressed versions.
    '''
    def __init__(self, fileobj):
        self._queue = queue.Queue() # XXX: we might limit queue-size (could cause deadlocks, though)
        self._fp = fileobj

        # patch original read-method
        self._orig_read = fileobj.read
        self._fp.read = self._tee_read

    def _tee_read(self, size=-1):
        buf = self._orig_read(size)
        self._queue.put(buf)

        return buf

    def iter_contents(self) -> collections.abc.Generator[bytes, None, None]:
        while (buf := self._queue.get()):
            yield buf


@contextlib.contextmanager
def tee_stream(
    fileobj: io.BytesIO,
    tee_receiver: collections.abc.Callable[[collections.abc.Generator[bytes, None, None]], None],
):
    '''
    ctx-mgr creating a "tee" for a given stream that will yield read chunks to the given
    `tee_receiver` callable. The passed-in stream patched and yielded back.

    The generator-receiving callback is executed in a thread, which will be joined when
    existing the managed ctx. As the thread will only terminate once the generator is
    exhausted (which in turn only happens if the stream is exhaused), the stream must
    be fully consumed before exiting the ctx.
    '''
    proxy = _TeeFilelikeProxy(fileobj=fileobj)

    def pass_tee_content_generator():
        tee_receiver(proxy.iter_contents())

    tee_receiver_thread = threading.Thread(
        target=pass_tee_content_generator,
        daemon=True,
    )

    try:
        tee_receiver_thread.start()
        yield fileobj
    finally:
        tee_receiver_thread.join()



import dataclasses
import functools
import json
import logging
import os
import os.path
import pkgutil
import sys
import threading
import typing
import urllib.parse

import dacite
import deprecated
import github3.exceptions
import yaml

import ci.log
import ctx

import model.base
from model.base import (
    ConfigElementNotFoundError,
    ModelValidationError,
    NamedModelElement,
)
from ci.util import (
    existing_dir,
    not_empty,
    not_none,
    parse_yaml_file,
)

dc = dataclasses.dataclass
empty_list = dataclasses.field(default_factory=list) # noqa:E3701
empty_tuple = dataclasses.field(default_factory=tuple) # noqa:E3701

ci.log.configure_default_logging()
logger = logging.getLogger(__name__)

'''
Configuration model and retrieval handling.

Users of this module will most likely want to create an instance of `ConfigFactory` and use it
to create `ConfigurationSet` instances.

Configuration sets are factories themselves, that are backed with a configuration source.
They create concrete configuration instances. While technically modifiable, all configuration
instances should not be altered by users. Configuration objects should usually not be
instantiated by users of this module.
'''


@dc(frozen=True)
class BaseConfigSetCfg:
    name: str


@dc(frozen=True)
class CfgTypeSrc: # just a marker class
    pass


@dc(frozen=True)
class LocalFileCfgSrc(CfgTypeSrc):
    file: str

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.file}'


@dc(frozen=True)
class GithubRepoFileSrc(CfgTypeSrc):
    repository_url: str
    relpath: str

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.repository_url} -> {self.relpath}'


@dc(frozen=True)
class ConfigTypeModel:
    factory_method: typing.Optional[str]
    cfg_type_name: str
    type: str


@dc(frozen=True)
class ConfigType:
    '''
    represents a configuration type (used for serialisation and deserialisation)
    '''
    model: ConfigTypeModel
    src: typing.Tuple[typing.Union[LocalFileCfgSrc, GithubRepoFileSrc], ...] = empty_tuple

    def sources(self):
        return self.src

    def factory_method(self):
        return self.model.factory_method

    def cfg_type_name(self):
        return self.model.cfg_type_name

    def cfg_type(self):
        return self.model.type


class ConfigFactory:
    '''Creates configuration model element instances from the underlying configuration source

    Configuration elements are organised in a two-level hierarchy: Configuration type
    (named cfg_type by convention) which specifies a configuration schema (and semantics) and
    Configuration element name (named cfg_name or element_name by convention).

    Configuration model elements may be retrieved through one of two methods:

        - via the generic `_cfg_element(cfg_type_name, cfg_name)`
        - via a "factory method" (if defined in cfg_type) - example: `github(cfg_name)`

    There is a special configuration type named `ConfigurationSet`, which is used to group
    sets of configuration elements. Configuration sets expose an API equivalent to ConfigFactory.
    '''

    CFG_TYPES = 'cfg_types'
    _cfg_type_cache = {} # <name>:<type>

    @staticmethod
    def _parse_local_file(cfg_dir: str, cfg_src: LocalFileCfgSrc):
        cfg_file = cfg_src.file
        return parse_yaml_file(os.path.join(cfg_dir, cfg_file))

    @staticmethod
    def _parse_repo_file(
        cfg_src: GithubRepoFileSrc,
        lookup_cfg_factory,
    ):
        import ccc.github
        repo_url = cfg_src.repository_url
        if not '://' in repo_url:
            repo_url = 'https://' + repo_url
        repo_url = urllib.parse.urlparse(repo_url)

        # shortcut lookup if local repo-mapping was configured
        if ctx.cfg and ctx.cfg.ctx and (repo_mappings := ctx.cfg.ctx.github_repo_mappings):
            for repo_mapping in repo_mappings:
                url = repo_mapping.repo_url
                if not '://' in url:
                    url = 'https://' + url
                url = urllib.parse.urlparse(url)
                if not url == repo_url:
                    continue
                fpath = os.path.join(repo_mapping.path, cfg_src.relpath)
                with open(fpath) as f:
                    return yaml.safe_load(f)

        if not lookup_cfg_factory:
            raise RuntimeError('cannot resolve non-local cfg w/o bootstrap-cfg-factory')

        gh_api = ccc.github.github_api(
            ccc.github.github_cfg_for_repo_url(
                repo_url,
                cfg_factory=lookup_cfg_factory,
            ),
            cfg_factory=lookup_cfg_factory,
        )
        org, repo = repo_url.path.strip('/').split('/')
        gh_repo = gh_api.repository(org, repo)

        try:
            file_contents = gh_repo.file_contents(
                path=cfg_src.relpath,
                ref=gh_repo.default_branch,
            ).decoded.decode('utf-8')

        except github3.exceptions.NotFoundError:
            logger.error(
                f"Unable to access file '{cfg_src.relpath}' in repository '{repo_url.path}'"
            )
            raise

        return yaml.safe_load(file_contents)

    @staticmethod
    def from_cfg_dir(
        cfg_dir: str,
        cfg_types_file='config_types.yaml',
        disable_cfg_element_lookup=False,
    ):
        if not cfg_dir:
            raise ValueError('cfg dir must not be None')

        cfg_dir = os.path.abspath(cfg_dir)
        if not os.path.isdir(cfg_dir):
            raise ValueError(f'{cfg_dir=} is not a directory')

        bootstrap_cfg_factory = ConfigFactory._from_cfg_dir(
            cfg_dir=cfg_dir,
            cfg_types_file=cfg_types_file,
            cfg_src_types=(LocalFileCfgSrc,),
            disable_cfg_element_lookup=False,
        )

        if disable_cfg_element_lookup:
            return bootstrap_cfg_factory

        return ConfigFactory._from_cfg_dir(
            cfg_dir=cfg_dir,
            cfg_types_file=cfg_types_file,
            cfg_src_types=None, # all
            lookup_cfg_factory=bootstrap_cfg_factory,
            disable_cfg_element_lookup=disable_cfg_element_lookup,
        )

    @staticmethod
    def _from_cfg_dir(
        cfg_dir: str,
        disable_cfg_element_lookup: bool,
        cfg_types_file='config_types.yaml',
        cfg_src_types=None,
        lookup_cfg_factory=None,
    ):
        cfg_dir = existing_dir(os.path.abspath(cfg_dir))
        cfg_types_dict = parse_yaml_file(os.path.join(cfg_dir, cfg_types_file))
        raw = {}

        raw[ConfigFactory.CFG_TYPES] = cfg_types_dict

        def retrieve_cfg(cfg_type):
            cfg_dict = {}

            for cfg_src in cfg_type.sources():
                if cfg_src_types and type(cfg_src) not in cfg_src_types:
                    continue

                if isinstance(cfg_src, LocalFileCfgSrc):
                    parsed_cfg = ConfigFactory._parse_local_file(
                        cfg_dir=cfg_dir,
                        cfg_src=cfg_src,
                    )
                elif isinstance(cfg_src, GithubRepoFileSrc):
                    if disable_cfg_element_lookup:
                        continue
                    parsed_cfg = ConfigFactory._parse_repo_file(
                        cfg_src=cfg_src,
                        lookup_cfg_factory=lookup_cfg_factory,
                    )
                else:
                    raise NotImplementedError(cfg_src)

                if parsed_cfg and parsed_cfg.items():
                    for k,v in parsed_cfg.items():
                        if k in cfg_dict and cfg_dict[k] != v:
                            raise ValueError(f'conflicting definition for {k=} in src {cfg_src}')
                        cfg_dict[k] = v

            return cfg_dict

        return ConfigFactory(
            raw_dict=raw,
            retrieve_cfg=retrieve_cfg,
        )

    @staticmethod
    def from_dict(raw_dict: dict):
        raw = not_none(raw_dict)

        return ConfigFactory(raw_dict=raw)

    def __init__(
        self,
        raw_dict: dict,
        retrieve_cfg: typing.Callable[[ConfigType], dict]=None,
    ):
        self.raw = not_none(raw_dict)
        if self.CFG_TYPES not in self.raw:
            raise ValueError(f'missing required attribute: {self.CFG_TYPES}')
        self.retrieve_cfg = retrieve_cfg

    def _retrieve_cfg_elements(self, cfg_type_name: str):
        if not cfg_type_name in self.raw:
            cfg_type = self._cfg_type(cfg_type_name=cfg_type_name)
            if self.retrieve_cfg:
                cfg_dict = self.retrieve_cfg(cfg_type)
            else:
                cfg_dict = {}
                # XXX hacky: use empty-dict if there is no retrieval-callable

            self.raw[cfg_type_name] = cfg_dict

    @deprecated.deprecated
    def _configs(self, cfg_name: str):
        '''
        returns all known cfg-element-names
        '''
        self._retrieve_cfg_elements(cfg_type_name=cfg_name)
        return self.raw[cfg_name]

    @functools.lru_cache
    def _cfg_types(self):
        return {
            cfg.cfg_type_name(): cfg for
            cfg in (
                dacite.from_dict(
                    data_class=ConfigType,
                    data=cfg_dict,
                    config=dacite.Config(
                        cast=[tuple],
                    ),
                ) for cfg_dict in self.raw[self.CFG_TYPES].values()
            )
        }

    def _cfg_type(self, cfg_type_name: str):
        self._ensure_type_is_known(cfg_type_name=cfg_type_name)
        return self._cfg_types().get(cfg_type_name)

    def _cfg_types_raw(self):
        return self.raw[self.CFG_TYPES]

    def cfg_set(self, cfg_name: str) -> 'ConfigurationSet':
        '''
        returns a new `ConfigurationSet` instance for the specified config name backed by the
        configured configuration source.
        '''
        configs_dict = self._configs('cfg_set')

        if cfg_name not in configs_dict:
            raise ValueError('no cfg named {c} in {cs}'.format(
                c=cfg_name,
                cs=', '.join(configs_dict.keys())
            )
            )
        return ConfigurationSet(
            cfg_factory=self,
            cfg_name=cfg_name,
            raw_dict=configs_dict[cfg_name]
        )

    def _cfg_element(self, cfg_type_name: str, cfg_name: str):
        cfg_type = self._cfg_type(cfg_type_name=cfg_type_name)

        # retrieve model class c'tor - search module and sub-modules
        # TODO: switch to fully-qualified type names
        own_module = sys.modules[__name__]

        submodule_names = [
            own_module.__name__ + '.' + m.name
            for m in pkgutil.iter_modules(own_module.__path__)
        ]
        for module_name in [__name__] + submodule_names:
            submodule_name = module_name.split('.')[-1]
            if module_name != __name__:
                module = getattr(__import__(module_name), submodule_name)
            else:
                module = sys.modules[submodule_name]

            # skip if module does not define our type
            if not hasattr(module, cfg_type.cfg_type()):
                continue

            # if type is defined, validate
            element_type = getattr(module, cfg_type.cfg_type())
            if not type(element_type) == type:
                raise ValueError()
            # found it (write to cache as part of crazy workaround for kaniko)
            self._cfg_type_cache[cfg_type_name] = element_type
            break
        else:
            # workaround for kaniko, which will purge our poor modules on multi-stage-builds
            if cfg_type_name in self._cfg_type_cache:
                element_type = self._cfg_type_cache[cfg_type_name]
            else:
                print(f'{self._cfg_type_cache=}')
                raise ValueError(f'failed to find cfg type: {cfg_type.cfg_type()=}')

        # for now, let's assume all of our model element types are subtypes of NamedModelElement
        # (with the exception of ConfigurationSet)
        configs = self._configs(cfg_type.cfg_type_name())
        if cfg_name not in configs:
            known_cfg_names = ', '.join(configs.keys())

            raise ConfigElementNotFoundError(
                f'cfg-factory: no such cfg-element: {cfg_name=} {cfg_type.cfg_type_name()=} '
                f'{known_cfg_names=}'
            )
        kwargs = {'raw_dict': configs[cfg_name]}

        if element_type == ConfigurationSet:
            kwargs.update({'cfg_name': cfg_name, 'cfg_factory': self})
        else:
            kwargs['name'] = cfg_name
            kwargs['type_name'] = cfg_type.cfg_type_name()

        element_instance = element_type(**kwargs)

        try:
            element_instance.validate()
        except ModelValidationError as mve:
            logger.warning(
                f"validation error for config '{cfg_name}' of type '{element_type.__name__}' "
                f"- ignored: {mve}"
            )

        return element_instance

    def _ensure_type_is_known(self, cfg_type_name: str):
        if cfg_type_name not in (known_types := self._cfg_types()):
            raise ValueError("Unknown config type '{c}'. Known types: {k}".format(
                c=cfg_type_name,
                k=', '.join(known_types.keys()),
            ))

    def _cfg_elements(self, cfg_type_name: str):
        '''Returns all cfg_elements for the given cfg_type.

        Parameters
        ----------
        cfg_type_name: str
            The name of the cfg_type whose instances should be retrieved.

        Yields
        -------
        NamedModelElement
            Instance of the given cfg_type.

        Raises
        ------
        ValueError
            If the specified cfg_type is unknown.
        '''
        not_empty(cfg_type_name)

        self._retrieve_cfg_elements(cfg_type_name=cfg_type_name)

        self._ensure_type_is_known(cfg_type_name=cfg_type_name)

        for element_name in self._cfg_element_names(cfg_type_name):
            yield self._cfg_element(cfg_type_name, element_name)

    def _cfg_element_names(self, cfg_type_name: str):
        '''Returns cfg-elements of the given cfg_type

        Parameters
        ----------
        cfg_type_name: str
            The cfg type name

        Returns
        -------
        Iterable[str]
            Contains the names of all cfg-elements of the given cfg_type known to this ConfigFactory.

        Raises
        ------
        ValueError
            If the specified cfg_type is unknown.
        '''
        not_empty(cfg_type_name)
        self._retrieve_cfg_elements(cfg_type_name=cfg_type_name)

        self._ensure_type_is_known(cfg_type_name=cfg_type_name)
        if cfg_type_name in self.raw:
            return set(self.raw[cfg_type_name].keys())
        else:
            return set()

    def __dir__(self):
        # prepend factory methods (improve REPL-shell experience)
        for cfg_type in self._cfg_types().values():
            if (factory_method := cfg_type.factory_method()):
                yield factory_method

            # also include methods for NamedModelElement-derived types
            elif cfg_type.cfg_type() == NamedModelElement.__name__:
                yield cfg_type.cfg_type_name()

        yield from super().__dir__()

    def __getattr__(self, cfg_type_name):
        for cfg_type in self._cfg_types().values():
            if cfg_type.factory_method() == cfg_type_name:
                return functools.partial(self._cfg_element, cfg_type_name)

            if cfg_type.cfg_type() == NamedModelElement.__name__:
                return functools.partial(self._cfg_element, cfg_type_name)

        raise AttributeError(cfg_type_name)

    def _serialise(self):
        cfg_types = self._cfg_types()

        serialised_elements = {
            cfg_name: self.retrieve_cfg(cfg_type) for cfg_name, cfg_type in cfg_types.items()
        }
        serialised_elements[ConfigFactory.CFG_TYPES] = self._cfg_types_raw()

        return json.dumps(serialised_elements, indent=2)

    def resolve_cfg_element_reference(
        self,
        cfg_element_reference: model.base.CfgElementReference,
    ):
        return self._cfg_element(
            cfg_type_name=cfg_element_reference.type_name,
            cfg_name=cfg_element_reference.element_name,
        )


class ConfigSetSerialiser:
    def __init__(self, cfg_sets: 'ConfigurationSet', cfg_factory: ConfigFactory):
        self.cfg_sets = not_none(cfg_sets)
        self.cfg_factory = not_none(cfg_factory)

    def serialise(self, output_format='json'):
        if not output_format == 'json':
            raise ValueError('not implemented')
        if len(self.cfg_sets) < 1:
            return '{}' # early exit for empty cfg_sets-set

        cfg_types = self.cfg_factory._cfg_types()
        # collect all cfg_names (<cfg-type>:[cfg-name])
        cfg_mappings = {}
        for cfg_mapping in [cfg_set._cfg_mappings() for cfg_set in self.cfg_sets]:
            for cfg_type_name, cfg_set_mapping in cfg_mapping:
                cfg_type = cfg_types[cfg_type_name]
                if cfg_type not in cfg_mappings:
                    cfg_mappings[cfg_type] = set()
                cfg_mappings[cfg_type].update(cfg_set_mapping['config_names'])

        # assumption: all cfg_sets share the same cfg_factory / all cfg_names are organized in one
        # global, flat namespace
        def serialise_element(cfg_type, cfg_names):
            elem_cfgs = {}

            for cfg_name in cfg_names:
                element = self.cfg_factory._cfg_element(cfg_type.cfg_type_name(), cfg_name)
                elem_cfgs[element.name()] = element.raw

            return (cfg_type.cfg_type_name(), elem_cfgs)

        serialised_elements = dict([serialise_element(t, n) for t, n in cfg_mappings.items()])

        # store cfg_set
        serialised_elements['cfg_set'] = {cfg.name(): cfg.raw for cfg in self.cfg_sets}

        # store cfg_types metadata (TODO: patch source attributes)
        serialised_elements[ConfigFactory.CFG_TYPES] = self.cfg_factory._cfg_types_raw()

        return json.dumps(serialised_elements, indent=2)


class ConfigurationSet(NamedModelElement):
    '''
    Represents a set of corresponding configuration. Instances are created by `ConfigFactory`.
    `ConfigurationSet` is itself a factory, offering a set of factory methods which create
    concrete configuration objects based on the backing configuration source.

    Note: This class is not thread safe until base_cfg_sets are resolved (_resolve_base_cfg_sets).

    Not intended to be instantiated by users of this module
    '''

    def __init__(self, cfg_factory, cfg_name, *args, **kwargs):
        self.resolved_base_cfg_sets = False
        self.cfg_factory: ConfigFactory = not_none(cfg_factory)
        self._raw_lock = threading.Lock()
        super().__init__(
            name=cfg_name,
            type_name='cfg_set',
            *args,
            **kwargs,
        )

        # normalise cfg mappings
        for cfg_type_name, entry in self.raw.items():
            if type(entry) == dict:
                entry = {
                    'config_names': entry['config_names'],
                    'default': entry.get('default', None)
                }
            elif type(entry) == str:
                entry = {'config_names': [entry], 'default': entry}

            self.raw[cfg_type_name] = entry

    def _base_cfgs(self):
        with self._raw_lock:
            return tuple(
                BaseConfigSetCfg(**cfg) for cfg in self.raw.get('_base_cfgs', {})
            )

    def _raw(self):
        return{
            k: v for k, v in self.raw.items() if not k == '_base_cfgs'
        }

    def _resolve_base_cfg_sets(self):
        if self.resolved_base_cfg_sets:
            return
        for base_cfg in self._base_cfgs():
            cfg_set = self.cfg_factory.cfg_set(cfg_name=base_cfg.name)
            self._merge_cfg_set(cfg_set=cfg_set, base_cfg=base_cfg)

        self.resolved_base_cfg_sets = True

    def _merge_cfg_set(self, cfg_set: 'ConfigurationSet', base_cfg: BaseConfigSetCfg):
        with self._raw_lock:
            cfg_set_dict = cfg_set._raw()

        for cfg_type_name, elements in cfg_set_dict.items():
            if not self.raw.get(cfg_type_name, {}):
                self.raw[cfg_type_name] = elements
            else:
                with self._raw_lock:
                    our_element_names = self._raw()[cfg_type_name]['config_names']

                # only add elements once
                other_element_names = [
                    cfg_name for cfg_name in elements.get('config_names')
                    if not cfg_name in our_element_names
                ]

                self.raw[cfg_type_name]['config_names'].extend(other_element_names)

                # use the first default from referenced based cfg sets if there is none
                if not self._raw()[cfg_type_name].get('default', None):
                    self.raw[cfg_type_name]['default'] = elements.get('default', None)

    def _optional_attributes(self):
        return {
            cfg_type_name for cfg_type_name in self.cfg_factory._cfg_types_raw()
        }

    def _cfg_mappings(self):
        self._resolve_base_cfg_sets()
        return self._raw().items()

    def _cfg_element(self, cfg_type_name: str, cfg_name=None):
        self.cfg_factory._ensure_type_is_known(cfg_type_name=cfg_type_name)

        cfg_name = self._default_name(cfg_type_name=cfg_type_name, cfg_name=cfg_name)

        return self.cfg_factory._cfg_element(
            cfg_type_name=cfg_type_name,
            cfg_name=cfg_name,
        )

    def _cfg_elements(self, cfg_type_name: str):
        '''Returns all container cfg elements of the given cfg_type

        Raises
        ------
        ValueError
            If the specified cfg_type is unknown.
        '''
        not_empty(cfg_type_name)

        for element_name in self._cfg_element_names(cfg_type_name):
            yield self._cfg_element(cfg_type_name, element_name)

    def _cfg_element_names(self, cfg_type_name: str):
        '''Returns all container cfg element names

        Raises
        ------
        ValueError
            If the specified cfg_type is unknown.
        '''
        not_empty(cfg_type_name)
        self._resolve_base_cfg_sets()
        # ask factory for all known names. This ensures that the existance of the type is checked.
        all_cfg_element_names = self.cfg_factory._cfg_element_names(cfg_type_name=cfg_type_name)

        if cfg_type_name in self._raw().keys():
            our_cfg_names = set(self._raw()[cfg_type_name]['config_names'])
            return (all_cfg_element_names & our_cfg_names) | our_cfg_names
        else:
            return set()

    def _default_name(self, cfg_type_name, cfg_name=None):
        if cfg_name:
            return cfg_name

        if not cfg_type_name in self._raw():
            self._resolve_base_cfg_sets()

        if not cfg_type_name in self._raw():
            raise ValueError(
                f'{self.name()=}: {cfg_type_name=} is unknown - known: {self._raw().keys()}'
            )

        cfg_name = self._raw()[cfg_type_name].get('default', None)
        if not cfg_name:
            self._resolve_base_cfg_sets()
            if not (cfg_name := self._raw()[cfg_type_name].get('default', None)):
                raise ValueError(f'No default for {cfg_type_name=}')
        return cfg_name

    def __getattr__(self, cfg_type_name):
        if not hasattr(self.cfg_factory, cfg_type_name):
            raise AttributeError(cfg_type_name)
        factory_method = getattr(self.cfg_factory, cfg_type_name)

        if not callable(factory_method):
            raise AttributeError(cfg_type_name)

        def get_default_element(cfg_name=None):
            if not cfg_name:
                cfg_name = self._default_name(cfg_type_name=cfg_type_name)

            return factory_method(cfg_name=cfg_name)
        return get_default_element

    def validate(self):
        cfg_types = self.cfg_factory._cfg_types()
        for cfg_type_name in cfg_types:
            for element in self._cfg_elements(cfg_type_name):
                element.validate()


def cluster_domain_from_kubernetes_config(cfg_factory, kubernetes_config_name: str):
    kubernetes_cfg = cfg_factory.kubernetes(kubernetes_config_name)
    if not (cluster_domain := kubernetes_cfg.cluster_domain()):
        raise ValueError(f'No cluster domain configured in {kubernetes_config_name=}')
    return cluster_domain



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import os
import typing

import dacite

import ocm

from ci.util import not_none
from concourse.client.model import (
    ResourceType,
)
from concourse.model.base import (
    AttributeSpec,
    ModelBase,
    ModelValidationError,
)
import model.github


def sane_env_var_name(name):
    not_none(name)
    return name.replace('-', '_').upper()


class ResourceIdentifier:
    def __init__(
        self,
        type_name,
        base_name,
        branch_name,
        qualifier=None,
        logical_name=None,
    ):
        self._type_name = not_none(type_name)
        self._base_name = not_none(base_name)
        self._branch_name = not_none(branch_name)
        self._qualifier = qualifier if qualifier else ''
        self._logical_name = logical_name

    def name(self):
        parts = [self._type_name, self._base_name, self._branch_name]
        if len(self._qualifier) > 0:
            parts.append(self._qualifier)

        return '-'.join(parts)

    def base_name(self):
        return self._base_name

    def branch_name(self) -> str:
        return self._branch_name

    def qualifier(self):
        return self._qualifier

    def type_name(self):
        return self._type_name

    def logical_name(self):
        return self._logical_name

    def __eq__(self, other):
        if not isinstance(other, ResourceIdentifier):
            return False
        return self.name() == other.name()

    def __hash__(self):
        return hash((self._type_name, self._base_name, self._qualifier))

    def __str__(self):
        return 'ResourceId: type {t}, base_name {bn}, qualifier {q}, resource_name {rn}'.format(
            t=self.type_name(),
            bn=self.base_name(),
            q=self._qualifier,
            rn=self.name()
        )


class Resource(ModelBase):
    def __init__(
        self,
        resource_identifier: ResourceIdentifier,
        *args,
        **kwargs
    ):
        self._resource_identifier = resource_identifier
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ()

    def resource_identifier(self):
        return self._resource_identifier

    def __str__(self):
        return 'Resource with id: {id}'.format(id=self._resource_identifier)

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return False
        return self._resource_identifier == other._resource_identifier

    def __hash__(self):
        return self._resource_identifier.__hash__()


class ResourceRegistry:
    def __init__(self):
        self.resources_dict = {}

    def __contains__(self, item):
        if isinstance(item, Resource):
            id = item.resource_identifier()
        elif isinstance(item, ResourceIdentifier):
            id = item
        else:
            return False
        return id in self.resources_dict

    def __getitem__(self, item):
        if isinstance(item, Resource):
            id = item.resource_identifier()
        elif isinstance(item, ResourceIdentifier):
            id = item
        else:
            id = item
        return self.resources_dict[id]

    def __len__(self):
        return len(self.resources_dict)

    def add_resource(self, resource, discard_duplicates=True):
        if not isinstance(resource, Resource):
            raise ValueError('not an instance of Resource')

        resource_id = resource.resource_identifier()
        if resource_id in self.resources_dict:
            if discard_duplicates:
                return # nothing to do (resource already existed)
            raise ValueError('insertion conflict: {id}'.format(id=resource_id))
        self.resources_dict[resource_id] = resource

    def resources(self, type_name, qualifier=None):
        def filter_expr(resource):
            id = resource.resource_identifier()
            if not id.type_name() == type_name:
                return False
            if qualifier is not None and id.qualifier() != qualifier:
                return False
            return True

        return filter(filter_expr, self.resources_dict.values())

    def resource(self, resource_identifier):
        return self[resource_identifier]


REPO_ATTRS = (
    AttributeSpec.optional(
        name='name',
        default='source',
        doc='''
        the logical repository name. affects environment variable names and is used to reference
        from traits.
        ''',
    ),
    AttributeSpec.optional(
        name='cfg_name',
        default=None,
        doc='''
        the github_cfg to use for authentication. Defaults to concourse-specific default
        ''',
    ),
    AttributeSpec.optional(
        name='force_push',
        default=False,
        doc='whether or not force-pushes ought to be done',
        type=bool,
    ),
    AttributeSpec.optional(
        name='trigger_paths',
        default={
            'include': [],
            'exclude': [],
        },
        doc='repository paths to either ignore or to restrict triggering to',
        type=dict,
    ),
    AttributeSpec.optional(
        name='trigger',
        default=None,
        doc='overwrites the defaults for triggering behaviour',
        type=bool,
    ),
    AttributeSpec.optional(
        name='disable_ci_skip',
        default=False,
        doc='whether to disable the ignoring of commits with [ci skip] in commit msg',
        type=bool,
    ),
    AttributeSpec.optional(
        name='branch',
        default=None,
        doc='only for non-main repository: specify branch to work with',
    ),
    AttributeSpec.optional(
        name='path',
        default=None,
        doc='github repository path (e.g. gardener/gardener)',
    ),
    AttributeSpec.optional(
        name='hostname',
        default=None,
        doc='do not use',
    ),
    AttributeSpec.optional(
        name='preferred_protocol',
        default=None,
        doc='optionally overwrites the preferred protocol to use',
        type=model.github.Protocol,
    ),
    AttributeSpec.optional(
        name='source_labels',
        default=[],
        type=typing.List[ocm.Label],
        doc='labels to add to the corresponding source declaration in base-component-descriptor'
    ),
    AttributeSpec.optional(
        name='submodule_cfg_names',
        default=[],
        type=list[str],
        doc='''
        The name of github configs to provide for authentication of submodules which are not stored
        on the same git server as the repository.
        The config used by the repository itself is included by default.
        The configs given must support http-auth, otherwise they will be omitted.
        '''
    )
)


class RepositoryConfig(Resource):
    def __init__(
            self,
            logical_name: str=None,
            qualifier: str=None,
            is_pull_request: bool=False,
            is_main_repo: bool=False,
            *args, **kwargs
        ):
        self._is_pull_request = is_pull_request
        self._is_main_repo = is_main_repo

        # todo: handle "qualifier"
        if is_pull_request:
            type_name = ResourceType.PULL_REQUEST.value
        else:
            type_name = ResourceType.GIT.value

        base_name = kwargs['raw_dict']['path'].replace('/', '.')

        branch_name = kwargs['raw_dict'].get('branch')

        resource_identifier = ResourceIdentifier(
            type_name=type_name,
            base_name=base_name,
            branch_name=branch_name,
            qualifier=qualifier,
            logical_name=logical_name
        )

        super().__init__(resource_identifier=resource_identifier, *args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return REPO_ATTRS

    def custom_init(self, raw_dict):
        if raw_dict.get('trigger') is not None:
            self._trigger = raw_dict['trigger']
        else:
            self._trigger = self._is_main_repo

        self._disable_ci_skip = raw_dict.get('disable_ci_skip', False)
        if 'disable_ci_skip' not in raw_dict:
            self._disable_ci_skip = not self._is_main_repo

    def cfg_name(self):
        return self.raw['cfg_name']

    def submodule_cfg_names(self) -> list[str]:
        return self.raw['submodule_cfg_names']

    def resource_name(self):
        # TODO: replace usages with access to resource_id
        return self._resource_identifier.name()

    def name(self):
        # TODO: replace usages with access to resource_id
        return self._resource_identifier.name()

    def logical_name(self):
        # TODO: replace usages with access to resource_id
        return self._resource_identifier.logical_name()

    def repo_path(self):
        return self.raw['path'] # owner/repo_name

    def repo_name(self):
        return self.repo_path().split('/')[-1]

    def repo_owner(self):
        return self.repo_path().split('/')[0]

    def repo_hostname(self):
        hostname = self.raw.get('hostname')
        if hostname is not None:
            return hostname.lower()
        return None

    def branch(self):
        return self.raw['branch']

    def should_trigger(self):
        return self._trigger

    def force_push(self):
        return self.raw['force_push']

    def _trigger_paths(self):
        return self.raw['trigger_paths']

    def trigger_include_paths(self):
        paths = self._trigger_paths()
        return paths['include']

    def trigger_exclude_paths(self):
        paths = self._trigger_paths()
        return paths['exclude']

    def source_labels(self):
        return [
            dacite.from_dict(
                data_class=ocm.Label,
                data=label_dict,
            ) for label_dict
            in self.source_label_dicts()
        ]

    def source_label_dicts(self):
        return self.raw['source_labels']

    def is_main_repo(self):
        return self._is_main_repo

    def disable_ci_skip(self):
        return self._disable_ci_skip

    def preferred_protocol(self):
        protocol_str = self.raw.get('preferred_protocol')
        if protocol_str:
            return model.github.Protocol(protocol_str)
        else:
            return None

    def head_sha_path(self):
        if self._is_pull_request:
            head_sha = '.git/head_sha'
        else:
            head_sha = '.git/HEAD'
        return os.path.join(self.resource_name(), head_sha)

    def pr_id_path(self):
        if not self._is_pull_request:
            raise RuntimeError('resource is not a pull-request')
        return os.path.join(self.resource_name(), '.git', 'id')

    def env_var_value_dict(self):
        name = self.logical_name()
        env_var_dict = {
            self.path_env_var_name(): self.resource_name(),
            sane_env_var_name(name) + '_BRANCH': self.branch(),
            sane_env_var_name(name) + '_GITHUB_REPO_OWNER_AND_NAME': self.repo_path(),
        }
        if self.is_main_repo():
            env_var_dict['MAIN_REPO_DIR'] = self.resource_name()
        return env_var_dict

    def path_env_var_name(self):
        return sane_env_var_name(self.logical_name() + '_PATH')

    def __str__(self):
        return 'RepositoryConfig ({cfg}:{rp}:{b})'.format(
            cfg=self.cfg_name() if self.cfg_name() else '<default>',
            rp=self.repo_path(),
            b=self.branch()
        )

    def validate(self):
        super().validate()

        try:
            for label in self.source_labels():
                pass # source_labels converts into ocm.Label
        except dacite.DaciteError as e:
            raise ModelValidationError(f'Invalid {label=}') from e # pylint: disable=E0601



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import sys
import os

# add modules from root dir to module search path
# so unit test modules can use regular imports
sys.path.extend(
    (
        os.path.join(
            os.path.realpath(os.path.dirname(__file__)),
            os.pardir,
            os.pardir,
            os.pardir,
            os.pardir
        ),
        os.path.realpath(os.path.dirname(__file__))
    )
)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import enum
import json
import logging
import typing

import requests
from ensure import ensure_annotations

from concourse.client.model import ResourceType
from http_requests import AuthenticatedRequestBuilder

from concourse.client.model import (
    Build,
    BuildEvents,
    BuildPlan,
    ConcourseJob,
    Pipeline,
    PipelineConfig,
    PipelineConfigResource,
    PipelineResource,
    ResourceVersion,
    SetPipelineResult,
    Worker,
)
from .routes import ConcourseApiRoutesBase, ConcourseApiRoutesV6_3_0

logger = logging.getLogger(__name__)

# Hard coded oauth user and password
# https://github.com/concourse/fly/blob/f4592bb32fe38f54018c2f9b1f30266713882c54/commands/login.go#L143
AUTH_TOKEN_REQUEST_USER = 'fly'
AUTH_TOKEN_REQUEST_PWD = 'Zmx5'


class ConcourseApiVersion(enum.Enum):
    V6_3_0 = '6.3.0'
    V6_5_1 = '6.5.1'


latest_concourse_api_version = ConcourseApiVersion.V6_5_1


class ConcourseApiFactory:
    '''Factory for ConcourseApi objects
    '''
    @staticmethod
    def create_api(
        base_url: str,
        team_name: str,
        verify_ssl: str,
        concourse_api_version: latest_concourse_api_version,
    ) -> 'ConcourseApiBase':
        # disable logging output for now (breaks template in lss)
        # logger.info(f'created client w/ {base_url=}')

        if not concourse_api_version:
            concourse_api_version = latest_concourse_api_version

        if concourse_api_version is ConcourseApiVersion.V6_5_1:
            routes = ConcourseApiRoutesV6_3_0(base_url=base_url, team=team_name)
            return ConcourseApiV6_5_1(
                routes=routes,
                verify_ssl=verify_ssl,
            )
        else:
            raise NotImplementedError(
                f'not supported: {concourse_api_version=}',
            )


def select_attr(name: str):
    return lambda o: o.get(name)


class ConcourseApiBase:
    '''
    Implements a subset of concourse REST API functionality.

    After creation, `login` ought to be invoked at least once to allow for the
    execution of requests that required autorization.

    @param base_url: concourse endpoint (e.g. https://ci.concourse.ci)
    @param team_name: the team name used for authentication
    @param verify_ssl: whether or not certificate validation is to be done
    '''

    AUTH_TOKEN_ATTRIBUTE_NAME = 'access_token'

    @ensure_annotations
    def __init__(
        self,
        routes: ConcourseApiRoutesBase,
        verify_ssl=False,
    ):
        self.routes = routes
        self.request_builder = None
        self.verify_ssl = verify_ssl

    @ensure_annotations
    def _get(self, url: str):
        if not self.request_builder:
            raise ValueError('login first')

        return self.request_builder.get(url, return_type='json')

    @ensure_annotations
    def _put(self, url: str, body: str, headers={}, use_auth_token=True):
        if not self.request_builder:
            raise ValueError('login first')

        return self.request_builder.put(url, body=body, headers=headers)

    @ensure_annotations
    def _post(self, url: str, body: str="", headers={}):
        if not self.request_builder:
            raise ValueError('login first')

        return self.request_builder.post(url, body=body, headers=headers)

    @ensure_annotations
    def _delete(self, url: str):
        if not self.request_builder:
            raise ValueError('login first')

        return self.request_builder.delete(url)

    def login(self, username: str, passwd: str):
        login_url = self.routes.login()
        form_data = {
            'grant_type': 'password',
            'password': passwd,
            'scope': 'openid profile email federated:id groups',
            'username': username,
        }

        response = requests.post(
            url=login_url,
            data=form_data,
            auth=(AUTH_TOKEN_REQUEST_USER, AUTH_TOKEN_REQUEST_PWD),
            headers={'content-type': 'application/x-www-form-urlencoded'},
            timeout=(4, 31),
        )
        response.raise_for_status()
        auth_token = response.json()[self.AUTH_TOKEN_ATTRIBUTE_NAME]
        self.request_builder = AuthenticatedRequestBuilder(
            auth_token=auth_token,
            verify_ssl=self.verify_ssl
        )
        return auth_token

    @ensure_annotations
    def set_pipeline(self, name: str, pipeline_definition):
        previous_version = self.pipeline_config_version(name)
        headers = {'x-concourse-config-version': previous_version}

        url = self.routes.pipeline_cfg(name)
        self._put(url, str(pipeline_definition), headers=headers)
        return SetPipelineResult.CREATED if previous_version is None else SetPipelineResult.UPDATED

    @ensure_annotations
    def delete_pipeline(self, name: str):
        url = self.routes.pipeline(pipeline_name=name)
        self._delete(url)

    @ensure_annotations
    def get_pipeline(self, name: str):
        url = self.routes.pipeline(pipeline_name=name)
        return Pipeline(self._get(url))

    def pipelines(self):
        pipelines_url = self.routes.pipelines()
        response = self._get(pipelines_url)
        return map(select_attr('name'), response)

    def order_pipelines(self, pipeline_names):
        url = self.routes.order_pipelines()
        self._put(url, json.dumps(pipeline_names))

    @ensure_annotations
    def pipeline_cfg(self, pipeline_name: str):
        pipeline_cfg_url = self.routes.pipeline_cfg(pipeline_name)
        response = self._get(pipeline_cfg_url)
        return PipelineConfig(response, concourse_api=self, name=pipeline_name)

    def pipeline_resources(
        self,
        pipeline_names: typing.Union[typing.Sequence[str], str],
        resource_type: typing.Optional[ResourceType]=None,
    ) -> typing.Generator[PipelineConfigResource, None, None]:
        if isinstance(pipeline_names, str):
            pipeline_names = [pipeline_names]

        resources = (self.pipeline_cfg(pipeline_name=name).resources for name in pipeline_names)
        for resource_list in resources:
            for resource in resource_list:
                resource: PipelineConfigResource
                if not resource_type or ResourceType(resource.type) is resource_type:
                    yield resource

    @ensure_annotations
    def pipeline_config_version(self, pipeline_name: str):
        pipeline_cfg_url = self.routes.pipeline_cfg(pipeline_name)
        response = self.request_builder.get(
                pipeline_cfg_url,
                return_type=None,
                check_http_code=False
        )
        if response.status_code == 404:
            return None # pipeline did not exist yet

        # ensure we did receive an error other than 404
        self.request_builder._check_http_code(response, pipeline_cfg_url)

        return response.headers['X-Concourse-Config-Version']

    @ensure_annotations
    def unpause_pipeline(self, pipeline_name: str):
        unpause_url = self.routes.unpause_pipeline(pipeline_name)
        self.request_builder.put(
                unpause_url,
                body=""
        )

    @ensure_annotations
    def pause_pipeline(self, pipeline_name: str):
        pause_url = self.routes.pause_pipeline(pipeline_name)
        self.request_builder.put(
                pause_url,
                body=""
        )

    @ensure_annotations
    def expose_pipeline(self, pipeline_name: str):
        expose_url = self.routes.expose_pipeline(pipeline_name)
        self.request_builder.put(
                expose_url,
                body="",
        )

    @ensure_annotations
    def job_builds(self, pipeline_name: str, job_name: str):
        '''
        Returns a list of Build objects for the specified job.
        The list is sorted by the build number, newest build last
        '''
        builds_url = self.routes.job_builds(pipeline_name, job_name)
        response = self._get(builds_url)
        builds = [Build(build_dict, self) for build_dict in response]
        builds = sorted(builds, key=lambda b: b.id())
        return builds

    @ensure_annotations
    def job_build(self, pipeline_name: str, job_name: str, build_name: str):
        build_url = self.routes.job_build(pipeline_name, job_name, build_name)
        response = self._get(build_url)
        return Build(response, self)

    @ensure_annotations
    def pause_job(self, pipeline_name: str, job_name: str):
        pause_url = self.routes.pause_job(pipeline_name, job_name)
        self.request_builder.put(
                pause_url,
                body=""
        )

    @ensure_annotations
    def unpause_job(self, pipeline_name: str, job_name: str):
        unpause_url = self.routes.unpause_job(pipeline_name, job_name)
        self.request_builder.put(
                unpause_url,
                body=""
        )

    @ensure_annotations
    def trigger_build(self, pipeline_name: str, job_name: str):
        trigger_url = self.routes.job_builds(pipeline_name, job_name)
        response = self._post(trigger_url)
        return Build(response.json(), self)

    @ensure_annotations
    def build_plan(self, build_id):
        build_plan_url = self.routes.build_plan(build_id)
        response = self._get(build_plan_url)
        return BuildPlan(response, self)

    @ensure_annotations
    def build_events(self, build_id):
        build_plan_url = self.routes.build_events(build_id)
        # TODO: this request never seems to send an "EOF"
        # (probably to support streaming)
        # --> properly handle this special case
        response = self.request_builder.get(
                build_plan_url,
                return_type=None,
                stream=True # passed to sseclient
        )
        return BuildEvents(response, self)

    @ensure_annotations
    def trigger_resource_check(self, pipeline_name: str, resource_name: str, retries:int=5):
        url = self.routes.resource_check(pipeline_name=pipeline_name, resource_name=resource_name)

        # Resource checks are triggered by a POST with an empty JSON-document as body against
        # the resource's check-url
        response = self.request_builder.post(url, check_http_code=False, body='{}', headers={})

        if response.ok:
            return

        if response.status_code == 500 and response.reason.startswith('parent type has no version'):
            if retries > 0:
                self.trigger_resource_check(pipeline_name, resource_name, retries-1)
            else:
                response.reason = (
                    f"Unable to check resource '{resource_name}' in pipeline "
                    f"'{pipeline_name}'. If this pipeline was recently deployed, please "
                    "try again in a minute."
                )
                response.raise_for_status()
        else:
            response.raise_for_status()

    @ensure_annotations
    def resource(self, pipeline_name: str, resource_name: str):
        url = self.routes.resource(pipeline_name, resource_name)
        response = self._get(url)
        return PipelineResource(response, self)

    @ensure_annotations
    def resource_versions(self, pipeline_name: str, resource_name: str):
        url = self.routes.resource_versions(pipeline_name=pipeline_name, resource_name=resource_name)
        response = self._get(url)
        if not response:
            return [] # new resources can have no versions
        return [ResourceVersion(raw=raw, concourse_api=None) for raw in response]

    @ensure_annotations
    def list_workers(self):
        url = self.routes.list_workers()
        workers_list = self._get(url)
        return [Worker(raw=worker, concourse_api=None) for worker in workers_list]

    @ensure_annotations
    def prune_worker(self, worker_name: str):
        url = self.routes.prune_worker(worker_name)
        self._put(url, "")

    @ensure_annotations
    def abort_build(self, build_id):
        url = self.routes.abort_build(build_id)
        self._put(url, "")

    @ensure_annotations
    def pin_resource_version(self, pipeline_name: str, resource_name: str, resource_version_id: int):
        url = self.routes.pin_resource_version(pipeline_name, resource_name, resource_version_id)
        self._put(url, "")

    @ensure_annotations
    def pin_and_comment_resource_version(
        self, pipeline_name: str, resource_name: str, resource_version_id: int, comment: str
    ):
        self.pin_resource_version(pipeline_name, resource_name, resource_version_id)
        self.pin_comment(pipeline_name, resource_name, comment)

    @ensure_annotations
    def unpin_resource(self, pipeline_name: str, resource_name: str) -> bool:
        resource = self.resource(pipeline_name, resource_name)
        if resource.is_pinned():
            url = self.routes.unpin_resource(pipeline_name, resource_name)
            self._put(url, "")
            return True
        return False

    @ensure_annotations
    def pin_comment(self, pipeline_name: str, resource_name: str, comment: str):
        url = self.routes.pin_comment(pipeline_name, resource_name)
        pin_comment = json.dumps({'pin_comment': comment})
        self._put(url, pin_comment)

    def get_job(self, pipeline_name:str, job_name:str):
        url = self.routes.job(pipeline_name, job_name)
        return ConcourseJob(self._get(url))

    def set_team(
        self,
        team_name: str,
        username: str,
        role: str='member',
        github_auth_teams: list[str]=[],
    ):
        body = {
            "auth": {
                role: {
                    "users": [
                        "local:" + username,
                    ]
                }
            }
        }
        if github_auth_teams:
            body["auth"][role].update({
                'groups': [f'github:{team}' for team in github_auth_teams]
            })

        team_url = self.routes.team_url(team_name)
        self._put(team_url, json.dumps(body))


class ConcourseApiV6_5_1(ConcourseApiBase):
    pass



import dataclasses
import logging
import re
import traceback
import typing

import ocm
import git
import github3.pulls

import cnudie.util
import cnudie.retrieve
import version

logger = logging.getLogger(__name__)

r'''
This pattern matches code-blocks in the following format:
```{category} {target_group} [source component name] [reference-dependent str] [author]
{note_message}
```
with the three groups in "[]" being optional by virtue of not being present for commit-attached
release note blocks.

Note: [^\S\n] is "all whitespaces except \n" (or "not [all non-whitespaces and newline]") to
approxmiate the as-of-yet unsupported \h ([:blank:]) aka "horizontal whitespace"

\x60 -> `
'''
_source_block_pattern = re.compile(
    pattern=(
        r'\x60{3}[^\S\n]*(?P<category>\w+)[^\S\n]+(?P<target_group>\w+)[^\S\n]*'
        r'(?P<source_component_name>\S+)?[^\S\n]?(?P<reference_str>\S+)'
        r'?[^\S\n]?(?P<author>\S+)?[^\S\n]*\n(?P<note>.+?)\n\x60{3}'
    ),
   flags=re.DOTALL | re.IGNORECASE | re.MULTILINE
)


@dataclasses.dataclass(frozen=True)
class Author:
    # for pull requests
    username: str  # the GitHub username

    # for commits
    display_name: str  # the name which authored the commit
    email: str  # the email the commit was authored

    def __str__(self) -> str:
        if self.username and self.username.strip():
            return f'@{self.username}'
        if self.display_name and self.display_name.strip():
            return f'`{self.display_name} <{self.email}>`'
        return ''


def author_from_commit(commit: git.Commit) -> Author:
    return Author(
        username='',
        display_name=commit.author.name,
        email=commit.author.email
    )


def author_from_pull_request(pull_request: github3.pulls.ShortPullRequest) -> Author:
    return Author(
        username=pull_request.user.login,
        display_name='',
        email=''
    )


def author_from_source(source: git.Commit | github3.pulls.ShortPullRequest) -> Author:
    if isinstance(source, git.Commit):
        return author_from_commit(source)
    elif isinstance(source, github3.pulls.ShortPullRequest):
        return author_from_pull_request(source)
    else:
        raise NotImplementedError(type(source))


@dataclasses.dataclass(frozen=True)
class _ReferenceType:
    identifier: str  # identifier for release note block
    prefix: str  # prefix for generated release notes


_ref_type_pull = _ReferenceType(identifier='#', prefix='#')
_ref_type_commit = _ReferenceType(identifier='$', prefix='@')

_ref_types = (_ref_type_pull, _ref_type_commit)


@dataclasses.dataclass(frozen=True)
class _Reference:
    ''' Represents where a release note comes from, for example through a
    commit or a pull request.

    _Reference is only a superclass for a commit- or pull request-reference,
    which have their own classes to access the pull request or the commit
    objects: `CommitReference` and `PullRequestReference`.  '''
    type: _ReferenceType

    @property
    def identifier(self) -> str:
        ''' The identifier for the reference - can be a commit hash, for
        example, or the number of the pull request.  '''
        raise NotImplementedError('get_content not implemented yet')


@dataclasses.dataclass(frozen=True)
class CommitReference(_Reference):
    ''' Represents the commit where the release note came from
    '''
    commit: git.Commit

    @property
    def identifier(self) -> str:
        return self.commit.hexsha


@dataclasses.dataclass(frozen=True)
class PullRequestReference(_Reference):
    ''' Represents the pull requests where the release note came from
    '''
    pull_request: github3.pulls.ShortPullRequest
    source_block: 'SourceBlock'

    @property
    def identifier(self) -> str:
        if self.source_block.reference_identifier:
            return self.source_block.reference_identifier.strip('#')
        return str(self.pull_request.number)


@dataclasses.dataclass(frozen=True)
class SourceBlock:
    '''Represents the parsed release note code block within a pull request body or a commit message.

    ```{category} {note_message} [component name] [reference identifier] [author]
    {note_message}
    ```
    with the triple of componentname, reference identifier, and author only being present for code-
    blocks being attached to pull requests
    '''
    source: object
    category: str
    target_group: str
    note_message: str
    component_name: str | None
    author: str | None
    reference_identifier: str | None

    @property
    def identifier(self) -> str:
        ''' returns a human-readable identifier which can be used e.g. for duplicate checking.
        does not include line breaks or spaces
        '''
        return f'[{self.category}<{self.target_group}>]{self.note_message}' \
            .lower().strip().replace(' ', '').replace('\n', '')

    def has_content(self) -> bool:
        ''' checks if there is any content in the source note block
        '''
        if self.note_message.strip().lower() == 'none':
            return False
        return all(z and z.strip() for z in (self.category, self.target_group, self.note_message))

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other):
        if isinstance(other, ReleaseNote):
            return self.__eq__(other.source_block)
        if isinstance(other, SourceBlock):
            return hash(other) == hash(self)
        return False


def create_commit_ref(commit: git.Commit, source_block: SourceBlock) -> CommitReference:
    return CommitReference(type=_ref_type_commit, commit=commit)


def create_pull_request_ref(
        pull_request: github3.pulls.ShortPullRequest,
        source_block: SourceBlock,
    ) -> PullRequestReference:
    return PullRequestReference(
        type=_ref_type_pull,
        pull_request=pull_request,
        source_block=source_block,
    )


def iter_source_blocks(source, content: str) -> typing.Generator[SourceBlock, None, None]:
    ''' Searches for code blocks in release note notation and returns all found.
    Only valid note blocks are returned, which means that the format has been followed.
    However, it does not check if the category / group exists.

    :param content: the content to look for release notes in
    :return: a list of valid note blocks
    '''
    while '<!--' in content:
        comment_start_idx = content.find('<!--')
        comment_stop_idx = content.find('-->', comment_start_idx + len('<!--'))
        content = content[:comment_start_idx] + content[comment_stop_idx + len('-->'):]

    for res in _source_block_pattern.finditer(content.replace('\r\n', '\n')):
        try:
            # component-name might be erroneously parsed if multiple values are specified
            # for target-group using pipe (|) character w/o spaces
            # as those are never valid OCM component-names, ignore those to avoid subsequent
            # processing errors (failure to retrieve component-descriptors/enumerate versions)
            component_name = res.group('source_component_name')
            if component_name and '|' in component_name:
                component_name = None

            block = SourceBlock(
                source=source,
                category=res.group('category'),
                target_group=res.group('target_group'),
                note_message=res.group('note'),
                author=res.group('author'),
                reference_identifier=res.group('reference_str'),
                component_name=component_name,
            )
            if not block.target_group.lower() in ['user', 'operator', 'developer', 'dependency']:
                continue
            if block.has_content():
                yield block
        except IndexError as e:
            logger.debug(f'cannot find group in content: {e}')
            # group not found, ignore
            continue


@dataclasses.dataclass(frozen=True)
class ReleaseNote:
    source_block: SourceBlock

    author: typing.Optional[Author]  # the author of the commit / pull request
    reference: _Reference

    source_component: ocm.Component
    is_current_repo: bool
    from_same_github_instance: bool

    def __hash__(self):
        return hash(self.source_block.identifier)

    def __eq__(self, other) -> bool:
        return self.source_block.__eq__(other)

    @property
    def reference_str(self) -> str:
        return f'{self.reference.type.identifier}{self.reference.identifier}'

    @property
    def block_str(self) -> str:
        src_blk = self.source_block
        author = (
            src_blk.author or self.author.username or self.author.display_name.replace(' ', '-')
        )
        if not author.startswith('@'):
            author = '@' + author
        return (
            f'```{src_blk.category} {src_blk.target_group} {self.source_component.name} '
            f'{self.reference_str} {author}\n'
            f'{src_blk.note_message}\n```'
        )


def _source_component(
    component_descriptor_lookup,
    version_lookup,
    current_component: ocm.Component,
    source_component_name: str,
) -> ocm.Component | None:
    try:
        # try to fetch greatest component-descriptor for source component. The
        # actual version (hopefully) does not matter, as we assume the GithubAccess
        # (which we need to lookup release-notes) will rarely change.
        source_component_descriptor = component_descriptor_lookup(
            ocm.ComponentIdentity(
                name=source_component_name,
                version=version.greatest_version(
                    versions=version_lookup(source_component_name),
                    ignore_prerelease_versions=True,
                ),
            ),
        )
        return source_component_descriptor.component
    except Exception:
        logger.warning(
            f'Unable to retrieve component descriptor for source component {source_component_name}'
        )
        traceback.print_exc()
        return None


def create_release_notes_obj(
    component_descriptor_lookup,
    version_lookup,
    source_block: SourceBlock,
    source_component: ocm.Component,
    current_component: ocm.Component,
) -> ReleaseNote:
    target = source_block.source
    if isinstance(target, git.Commit):
        ref = create_commit_ref(target, source_block)
    elif isinstance(target, github3.pulls.ShortPullRequest):
        ref = create_pull_request_ref(target, source_block)
    else:
        raise NotImplementedError(
            f"Release note creation not implemented for target-type {type(target)}"
        )

    author = author_from_source(target)

    if source_block.component_name:
        source_component = _source_component(
            component_descriptor_lookup=component_descriptor_lookup,
            version_lookup=version_lookup,
            current_component=current_component,
            source_component_name=source_block.component_name,
        )

    # access
    source_component_access = cnudie.util.determine_main_source_for_component(
        component=source_component,
        absent_ok=False
    ).access

    current_src_access = cnudie.util.determine_main_source_for_component(
        component=current_component,
        absent_ok=False
    ).access

    from_same_github_instance = current_src_access.hostname() in source_component_access.hostname()
    is_current_repo = current_component.name == source_component.name

    return ReleaseNote(
        source_block=source_block,
        author=author,
        reference=ref,
        source_component=source_component,
        is_current_repo=is_current_repo,
        from_same_github_instance=from_same_github_instance
    )


@dataclasses.dataclass(frozen=True)
class ReleaseNotesMetadata:
    checked_at: int
    prs: list[int]


@dataclasses.dataclass(frozen=True)
class MetaPayload:
    type: str
    data: object


def get_meta_obj(
        typ: str,
        data: object
) -> dict:
    return {
        'meta': dataclasses.asdict(MetaPayload(typ, data))
    }



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import enum
import textwrap
import typing

import dacite

from ci.util import not_none
from ocm import Label
import ci.util
import cnudie.retrieve
import ocm
import model.base
import version

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
    StepNotificationPolicy,
)
from concourse.model.base import (
    AttributeSpec,
    ModelValidationError,
    ScriptType,
    Trait,
    TraitTransformer,
)
import model.ctx_repository

OcmRepositoryMappingEntry = cnudie.retrieve.OcmRepositoryMappingEntry


@dataclasses.dataclass(frozen=True)
class StepInput:
    step_name: str
    output_name: str = None # if absent, use only output
    type: str = 'step'


class UploadMode(enum.StrEnum):
    LEGACY = 'legacy'
    NO_UPLOAD = 'no-upload'


DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME = 'component_descriptor'

ATTRIBUTES = (
    AttributeSpec.optional(
        name='step',
        default={'name': DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME},
        doc='The build step name injected by this trait',
        type=dict,
    ),
    AttributeSpec.optional(
        name='resolve_dependencies',
        default=True,
        doc='Indicates whether or not unresolved component dependencies should be resolved',
        type=bool,
    ),
    AttributeSpec.optional(
        name='upload',
        default='legacy',
        doc=textwrap.dedent('''\
            Indicates whether or not to publish component-descriptor during Component-Descriptor
            step. For backwards-compatibility reasons, defaults to "legacy", which is a mode
            that depends on pipeline's context.
            Should be configured to `no-upload` (which is the intended default behaviour)
        '''),
        type=UploadMode,
    ),
    AttributeSpec.optional(
        name='component_name',
        default=None, # actually, it is determined at runtime
        doc='Manually overwrites the component name (which defaults to github repository path)',
    ),
    AttributeSpec.optional(
        name='callback_env',
        default={},
        doc='Specifies additional environment variables passed to .ci/component_descriptor script',
    ),
    AttributeSpec.optional(
        name='retention_policy',
        default=None,
        type=typing.Union[version.VersionRetentionPolicies, str],
        doc='''
            specifies how (old) component-descriptors and their referenced resources should be
            handled. This is foremostly intended as an option for automated cleanup for
            components with frequent (shortlived) releases and/or frequent (shortlived) snapshots.

            if no retention_policy is defined, no cleanup will be done.

            retention policy may either be defined "inline" (as a mapping value) or by referencing
            a pre-defined policy by name (see `retentions_policies` attribute). In the latter case,
            use the policie's name as (string) attribute value.
        ''',
    ),
    AttributeSpec.optional(
        name='retention_policies',
        default=[
            version.VersionRetentionPolicies(
                name='clean-snapshots',
                rules=[
                    version.VersionRetentionPolicy(
                        name='clean-snapshots',
                        keep=64,
                        match=version.VersionType.SNAPSHOT,
                    ),
                    version.VersionRetentionPolicy(
                        name='keep-releases',
                        keep='all',
                        match=version.VersionType.RELEASE,
                    ),
                ],
                dry_run=False,
            ),
            version.VersionRetentionPolicies(
                name='clean-snapshots-and-releases',
                rules=[
                    version.VersionRetentionPolicy(
                        name='clean-snapshots',
                        keep=64,
                        match=version.VersionType.SNAPSHOT,
                    ),
                    version.VersionRetentionPolicy(
                        name='clean-releases',
                        keep=128,
                        match=version.VersionType.RELEASE,
                    ),
                ],
                dry_run=False,
            ),
        ],
        type=typing.List[version.VersionRetentionPolicies],
        doc='''
            predefined retention policies (see default value). may be referenced via
            `retention_policy` attribute (adding additional policies here has no immediate effect)
        '''
    ),
    AttributeSpec.deprecated(
        name='validation_policies',
        type=typing.List[str],
        default=['ignore-me'],
        doc='obsolete',
    ),
    AttributeSpec.deprecated(
        name='ctx_repository_base_url',
        type=str,
        default=None, # if not explicitly configured, will be injected from cicd-default
        doc='''
            the component descriptor context repository base_url (for component descriptor v2).
            If not configured, the CICD-landscape's default ctx will be used.
        '''
    ),
    AttributeSpec.deprecated(
        name='ctx_repository',
        type=str,
        default=None, # if not explicitly configured, will be injected from cicd-default
        doc='''
            the component descriptor context repository cfg name (for component descriptor v2).
            If not configured, the CICD-landscape's default ctx will be used.
        '''
    ),
    AttributeSpec.optional(
        name='ocm_repository',
        type=str,
        default=None, # if not explicitly configured, will be injected from cicd-default
        doc='''\
            the Component Descriptor OCM Repository url used for publishing.
            instead of repository-url, may also be the name of a cfg-element.
        '''
    ),
    AttributeSpec.optional(
        name='component_labels',
        default=[],
        type=typing.List[Label],
        doc='a list of labels to add to the component in the base Component Descriptor',
    ),
    AttributeSpec.optional(
        name='inputs',
        default=[],
        type=typing.List[StepInput],
        doc='inputs to expose to component-descriptor step',
    ),
    AttributeSpec.optional(
        name='depends',
        default=[],
        type=list[str],
        doc='steps that need to be run prior to running component-descriptor-step',
    ),
    AttributeSpec.optional(
        name='ocm_repository_mappings',
        default=[], # cannot define a proper default here because this depends on another (optional)
                    # config-value. At least not in a way that would be represented in our
                    # rendered documentation.
        type=list[OcmRepositoryMappingEntry],
        doc='''\
            used to explicitly configure where to lookup component descriptors. If given,
            ocm_repository *must* be defined.

            **Example:**

            .. code-block:: yaml

                - repository: ocm_repo_url
                  prefix: github.com/some-org/
                - repository: ocm_repo_url
                  prefix: github.com/another-org/
                - repository: another_ocm_repo_url
                  prefix: github.com/yet-another-org/

            Value from ocm_repository is implicitly added as entry.
        '''
    ),
)


class ComponentDescriptorTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # todo: make step name actually configurable (need concept to express
        # step-specific behaviour, first)
        if not self.step_name() == DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME:
            raise ModelValidationError(
                f"component-descriptor step name must be '{DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME}'"
            )

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def component_name(self):
        return self.raw['component_name']

    def step_name(self):
        return self.raw['step']['name']

    @property
    def upload(self) -> UploadMode:
        return UploadMode(self.raw['upload'])

    def retention_policy(self, raw=True) -> version.VersionRetentionPolicies | None:
        if not (policy := self.raw.get('retention_policy', None)):
            return None

        if isinstance(policy, str):
            # lookup name
            for candidate in self.raw.get('retention_policies', ()):
                if isinstance(candidate, dict):
                    name = candidate['name']
                elif isinstance(candidate, version.VersionRetentionPolicies):
                    name = candidate.name
                else:
                    raise ValueError(candidate)

                if name == policy:
                    policy = candidate
                    break
            else:
                raise ValueError(f'did not find {policy=} in retention_policies')

        if raw:
            if isinstance(policy, version.VersionRetentionPolicies):
                policy = dataclasses.asdict(
                    policy,
                    dict_factory=ci.util.dict_factory_enum_serialisiation,
                )
            return policy

        if isinstance(policy, version.VersionRetentionPolicies):
            return policy

        return dacite.from_dict(
            data_class=version.VersionRetentionPolicies,
            data=policy,
            config=dacite.Config(
                cast=(enum.Enum,),
            ),
        )

    def resolve_dependencies(self):
        return self.raw['resolve_dependencies']

    def callback_env(self) -> dict:
        return self.raw['callback_env']

    def validation_policies(self):
        return ()

    @property
    def ocm_repository(self) -> ocm.OciOcmRepository:
        ocm_repo = self.raw.get('ocm_repository') or self.raw.get('ctx_repository')
        # XXX hack for unittests
        if not self.cfg_set:
            return None

        # ocm_repo might be:
        # - a cfg-element-name (type: CtxRepositoryCfg)
        # - a ocm-repository-url (type: str)
        # - a OCM Repository (type: dict / gcm.OcmRepository

        if ocm_repo:
            try:
                ctx_repo_cfg = self.cfg_set.ctx_repository(ocm_repo)
            except model.base.ConfigElementNotFoundError:
                # assume it is a ocm-repository-url
                return ocm.OciOcmRepository(
                    baseUrl=ocm_repo,
                )
        else:
            ctx_repo_cfg = self.cfg_set.ctx_repository()

        ctx_repo_cfg: model.ctx_repository.CtxRepositoryCfg

        return ocm.OciOcmRepository(
            baseUrl=ctx_repo_cfg.base_url(),
        )

    def ctx_repository_base_url(self):
        ctx_repo = self.ocm_repository
        # XXX hack for unittsts
        if ctx_repo is None:
            return None

        # use default ctx_repository_base_url, if not explicitly configured
        if not (base_url := self.raw.get('ctx_repository_base_url')):
            return ctx_repo.baseUrl
        else:
            # XXX warn or even forbid, at least if different from ctx-repo-cfg?
            return base_url

    def component_labels(self):
        return self.raw['component_labels']

    def ocm_repository_mappings(self) -> list[OcmRepositoryMappingEntry]:
        ocm_repository_url = self.ctx_repository_base_url()
        ocm_repository_mappings: list[dict] = self.raw['ocm_repository_mappings']

        if not ocm_repository_url and not ocm_repository_mappings:
            return []

        mappings = [
            dacite.from_dict(
                data_class=OcmRepositoryMappingEntry,
                data=raw_mapping,
            ) for raw_mapping in ocm_repository_mappings
        ]

        if ocm_repository_url:
            for mapping in mappings:
                if mapping.repository == ocm_repository_url:
                    break
            else:
                mappings.append(
                    OcmRepositoryMappingEntry(
                        repository=ocm_repository_url,
                    )
                )

        return mappings

    def inputs(self) -> typing.List[StepInput]:
        return [
            dacite.from_dict(data_class=StepInput, data=raw_input)
            for raw_input in self.raw['inputs']
        ]

    @property
    def depends(self) -> list[str]:
        return self.raw['depends']

    def transformer(self):
        return ComponentDescriptorTraitTransformer(trait=self)

    def validate(self):
        super().validate()

        ocm_repository_url = self.ctx_repository_base_url()
        ocm_repository_mappings: list[dict] = self.raw['ocm_repository_mappings']
        if not ocm_repository_url and ocm_repository_mappings:
            raise ModelValidationError(
                'ocm_repository_url must be defined if ocm_repository_mappings are'
            )

        for label in self.component_labels():
            try:
                dacite.from_dict(
                    data_class=Label,
                    data=label,
                    config=dacite.Config(strict=True),
                )
            except dacite.DaciteError as e:
                raise ModelValidationError(
                    f"Invalid label '{label}'."
                ) from e


DIR_NAME = 'component_descriptor_dir'
ENV_VAR_NAME = 'component_descriptor_dir'


class ComponentDescriptorTraitTransformer(TraitTransformer):
    name = 'component_descriptor'

    def __init__(self, trait: ComponentDescriptorTrait, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trait = not_none(trait)

    def inject_steps(self):
        self.descriptor_step = PipelineStep(
            name=self.trait.step_name(),
            raw_dict={},
            is_synthetic=True,
            notification_policy=StepNotificationPolicy.NO_NOTIFICATION,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
        )
        self.descriptor_step.add_output(
            name=DIR_NAME,
            variable_name=ENV_VAR_NAME,
        )
        self.descriptor_step.set_timeout(duration_string='45m')

        yield self.descriptor_step

    def process_pipeline_args(self, pipeline_args: 'JobVariant'):
        if pipeline_args.has_step('release'):
            release_step = pipeline_args.step('release')
            release_step.add_input(
                name=DIR_NAME,
                variable_name=ENV_VAR_NAME,
            )
        if pipeline_args.has_trait('draft_release'):
            draft_release_step = pipeline_args.step('create_draft_release_notes')
            draft_release_step.add_input(
                name=DIR_NAME,
                variable_name=ENV_VAR_NAME,
            )

        # inject component_name if not configured
        if not self.trait.raw.get('component_name'):
            main_repo = pipeline_args.main_repository()
            component_name = '/'.join((
                main_repo.repo_hostname(),
                main_repo.repo_path(),
            ))
            self.trait.raw['component_name'] = component_name

        for step_name_to_depend_on in self.trait.depends:
            step_to_depend_on = pipeline_args.step(step_name_to_depend_on)
            self.descriptor_step._add_dependency(step_to_depend_on)

            # avoid cyclic dependencies (user-defined depends should have precedence)
            if self.descriptor_step.name in step_to_depend_on.depends():
                step_to_depend_on._remove_dependency(self.descriptor_step)

        # add configured (step-)inputs
        for step_input in self.trait.inputs():
            if not step_input.type == 'step':
                raise NotImplementedError(step_input.type)

            try:
                step: PipelineStep = pipeline_args.step(step_input.step_name)
            except KeyError as ke:
                raise ValueError(f'no such step: {step_input.step_name=}') from ke

            self.descriptor_step._add_dependency(step)

            if step_input.output_name:
                output_name = step_input.output_name
            else:
                # choose only output if omitted
                outputs = {
                    name: v for name,v in step.outputs().items()
                    if not name == 'on_error_dir' # XXX hack hack hack
                }
                if len(outputs) < 1:
                    raise ValueError(f'{step.name=} does not have any outputs')
                elif len(outputs) > 1:
                    raise ValueError(
                        f'{step.name=} has more than one output (need to tell step_name)'
                    )
                output_name = next(outputs.keys().__iter__())

            self.descriptor_step.add_input(
                name=output_name,
                variable_name=output_name,
            )

    @classmethod
    def dependencies(cls):
        return {'version'}

    @classmethod
    def order_dependencies(cls):
        # dependency is required, as we need to patch the 'release' step
        return {'release'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import concourse.model.resources


class PipelineDefinition:
    def __init__(self):
        self._variants_dict = {}
        self._resource_registry = concourse.model.resources.ResourceRegistry()

    def resource_registry(self):
        return self._resource_registry

    def variants(self):
        return self._variants_dict.values()

    def variant(self, name: str):
        return self._variants_dict[name]



import ci.util
import logging
import os

import ocm

logger: logging.Logger = logging.getLogger(__name__)


def component_descriptor_fname(
    schema_version=ocm.SchemaVersion.V2,
):
    schema_version = ocm.SchemaVersion(schema_version)
    if schema_version is ocm.SchemaVersion.V2:
      return 'component_descriptor_v2'
    else:
        raise NotImplementedError(schema_version)


def component_descriptor_path(
    schema_version=ocm.SchemaVersion.V2,
):
    fname = component_descriptor_fname(schema_version=schema_version)
    return os.path.join(
      ci.util.check_env('COMPONENT_DESCRIPTOR_DIR'),
      fname,
    )


def parse_component_descriptor(
    schema_version=ocm.SchemaVersion.V2,
):
    if schema_version is ocm.SchemaVersion.V2:
      component_descriptor = ocm.ComponentDescriptor.from_dict(
          component_descriptor_dict=ci.util.parse_yaml_file(
            component_descriptor_path(
              schema_version=schema_version,
            )
          )
      )
      return component_descriptor
    else:
        raise NotImplementedError(schema_version)


def component_descriptor_from_component_descriptor_path(
    cd_path: str,
) -> ocm.ComponentDescriptor:

    if not os.path.isfile(cd_path):
        raise FileNotFoundError(
            f'{os.path.abspath(cd_path)=} not found'
        )

    descriptor_v2 = ocm.ComponentDescriptor.from_dict(
        ci.util.parse_yaml_file(cd_path)
    )
    logger.info(f'found component-descriptor (v2) at {cd_path=}')
    return descriptor_v2


def component_descriptor_from_dir(dir_path: str) -> ocm.ComponentDescriptor:
    if not os.path.isdir(dir_path):
        raise NotADirectoryError(
            f'{os.path.abspath(dir_path)=} is no directory'
        )

    component_descriptor_path = os.path.join(
        dir_path,
        component_descriptor_fname(
            schema_version=ocm.SchemaVersion.V2,
        ),
    )

    have_cd = os.path.exists(component_descriptor_path)

    if have_cd:
        return component_descriptor_from_component_descriptor_path(
            cd_path=component_descriptor_path,
        )
    else:
        print(f'did not find expected component-descriptor at {component_descriptor_path=}')
        exit(1)
